#! /usr/bin/env python
import unittest
from datetime import datetime, timedelta
from pyschism.forcing import Tides


class TidesTestCase(unittest.TestCase):

    def setUp(self):
        self.tf = Tides()
        self.tf.start_date = datetime(2017, 9, 23)
        self.tf.end_date = datetime(2017, 9, 25)
        self.tf.spinup_time = timedelta(days=7)
        self.tf.use_constituent('M2')

    def test_use_ll(self):
        self.tf.use_all()

    def test___call__(self):
        self.tf('M2')

    def test___iter__(self):
        for c, v in self.tf:
            pass

    def test___len__(self):
        len(self.tf)

    def test_use_major(self):
        self.tf.use_major()

    def test_drop_constituent(self):
        self.tf.drop_constituent('M2')

    def test_get_active_constituents(self):
        self.tf.get_active_constituents()

    def test_get_nodal_factor(self):
        for f in self.tf.orbital_frequencies:
            self.tf.get_nodal_factor(f)

    def test_get_greenwich_factor(self):
        for f in self.tf.orbital_frequencies:
            self.tf.get_greenwich_factor(f)

    def test_get_nodal_factor_raise(self):
        self.assertRaises(TypeError, self.tf.get_nodal_factor, '')

    def test_get_greenwich_factor_raise(self):
        self.assertRaises(TypeError, self.tf.get_greenwich_factor, '')

    def test_reset_dates(self):
        new_date = self.tf.start_date
        del(self.tf.start_date)
        self.tf.end_date = new_date

    def test_use_all(self):
        self.tf.use_all()

    def test_deafault_spiunp_time(self):
        del(self.tf.spinup_time)
        self.tf.spinup_time


if __name__ == '__main__':
    unittest.main()



from datetime import datetime, timedelta
import pathlib
from typing import Union

import pytz

from pyschism.forcing.tides.tides import Tides
from pyschism.mesh import Hgrid, hgrid

def datetime_is_naive(d) -> bool:
    return d.tzinfo is None or d.tzinfo.utcoffset(d) is None

def datetime_is_aware(d) -> bool:
    return d.tzinfo is not None and d.tzinfo.utcoffset(d) is not None


class HgridDescriptor:

    def __set__(self, obj, val: Hgrid):
        if not isinstance(val, Hgrid):
            raise TypeError('Argument hgrid must be of type {Hgrid}, not type '
                            f'{type(val)}.')
        obj.__dict__['hgrid'] = val

    def __get__(self, obj, val):
        return obj.__dict__['hgrid']

class StartDateDescriptor:

    def __set__(self, obj, val: datetime):
        if not isinstance(val, datetime):
            raise TypeError(
                    f'Argument start_date must be of type {datetime}, '
                    f'not type {type(val)}.')
        if datetime_is_naive(val):
            val = pytz.timezone('UTC').localize(val)
        obj.__dict__['start_date'] = val

    def __get__(self, obj, val):
        return obj.__dict__['start_date']
        
class RndayDescriptor:

    def __set__(self, obj, val: Union[int, float, timedelta]):
        if not isinstance(val, (int, float, timedelta)):
            raise TypeError(
                f'Argument rnday must be of type {int}, {float} or '
                f'{timedelta}, not type {type(val)}.')
        if not isinstance(val, timedelta):
            val = timedelta(days=val)
        obj.__dict__['rnday'] = val 

    def __get__(self, obj, val) -> timedelta:
        return obj.__dict__['rnday']

class Bctides:

    _hgrid = HgridDescriptor()
    _start_date = StartDateDescriptor()
    _rnday = RndayDescriptor()

    def __init__(
            self,
            hgrid: Hgrid,
            start_date: datetime,
            rnday: timedelta,
            cutoff_depth=50.,
            elevation=True,
            velocity=False,
            temperature=False,
            salinity=False,
            tracers=False,
            tidal_database='hamtide',
            # ts_database='hycom',
    ):
        self._hgrid = hgrid
        self._start_date = start_date
        self._rnday = rnday
        self._cutoff_depth = cutoff_depth
        self._elevation = elevation
        self._velocity = velocity
        self._temperature = temperature
        self._salinity = salinity
        self._tracers = [] if tracers is None or tracers is False else tracers
        self._tides = Tides(database=tidal_database)
        self.tides.use_all()

    def __str__(self):
        f = [
            f'{str(self.start_date)}',
            f'{self.ntip} {self.cutoff_depth}',
        ]
        if self.ntip > 0:
            for constituent in self.tides.get_active_potential_constituents():
                forcing = self.tides(
                    self.start_date, self.rnday, constituent)
                f.append(' '.join([
                    f'{constituent}\n',
                    f'{forcing[0]:G}',
                    f'{forcing[1]:G}',
                    f'{forcing[2]:G}',
                    f'{forcing[3]:G}',
                    f'{forcing[4]:G}']))
        f.append(f'{self.tides.nbfr:d}')
        for constituent in self.tides.get_active_forcing_constituents():
            forcing = self.tides(
                self.start_date, self.rnday, constituent)
            f.append(' '.join([
                f'{constituent}\n',
                f'{forcing[2]:G}',
                f'{forcing[3]:G}',
                f'{forcing[4]:G}']))
        f.append(f'{len(self.hgrid.boundaries.ocean())}')
        for boundary in self.hgrid.boundaries.ocean().itertuples():
            f.append(self.get_forcing(boundary))
        return '\n'.join(f)

    def write(self, path, overwrite: bool = False):
        path = pathlib.Path(path)
        if path.exists() and not overwrite:
            raise IOError('path exists and overwrite is False')
        open(path, 'w').write(str(self))

    def get_forcing(self, boundary):

        def forcing_digit(pos):
            if pos == 0:
                return f'{len(boundary.indexes)}'

            elif pos == 1:
                if self.elevation is True:
                    return '3'
                elif self.elevation is False or self.elevation is None:
                    return '0'
                else:
                    raise TypeError('Unhandled elevation source type '
                                    f'{type(self.elevation)}.')

            elif pos == 2:
                if self.velocity is True:
                    return '3'
                elif self.velocity is False or self.velocity is None:
                    return '0'
                else:
                    raise TypeError('Unhandled velocity source type '
                                    f'{type(self.velocity)}.')

            elif pos == 3:
                if self.temperature is False or self.temperature is None:
                    return '0'
                else:
                    raise TypeError('Unhandled temperature source type '
                                    f'{type(self.temperature)}.')

            elif pos == 4:
                if self.salinity is False or self.salinity is None:
                    return '0'
                else:
                    raise TypeError('Unhandled salinity source type '
                                    f'{type(self.salinity)}.')

            else:
                raise ValueError(f'Unhandled argument pos={pos}.')
        
        def get_forcing(pos, digit):
            f =[]
            if pos == 1: # elevation
                if digit == 3: # tides
                    if self.elevation is True: # default tidal elevation
                        for constituent in \
                                self.tides.get_active_forcing_constituents():
                            f.append(f'{constituent}')
                            vertices = self.hgrid.get_xy(
                                crs='EPSG:4326')[boundary.indexes, :]
                            amp, phase = self.tides.get_elevation(
                                constituent, vertices)
                            for i in range(len(amp)):
                                f.append(f'{amp[i]:.8e} {phase[i]:.8e}')
                        return '\n'.join(f)
                    else:
                        raise TypeError(
                            f'Unhandled elevation type {type(self.elevation)}')
                else:
                    raise ValueError(
                        f'Unhandled elevation forcing digit={digit}')
            elif pos == 2: # velocity
                if self.velocity is True: # default tidal velocity
                    for constituent in \
                            self.tides.get_active_forcing_constituents():
                        f.append(f'{constituent}')
                        vertices = self.hgrid.get_xy(
                            crs='EPSG:4326')[boundary.indexes, :]
                        uamp, uphase, vamp, vphase = self.tides.get_velocity(
                            constituent, vertices)
                        for i in range(len(vertices)):
                            f.append(f'{uamp[i]:.8e} {uphase[i]:.8e} ' \
                                        f'{vamp[i]:.8e} {vphase[i]:.8e}')
                    return '\n'.join(f)
                else:
                    raise TypeError(
                        f'Unhandled velocity type {type(self.velocity)}')
            else:
                raise TypeError('Unhandled forcing at pos={pos}.')


        f = [' '.join(list(map(forcing_digit, range(5 + len(self._tracers)))))]
        for i in range(1, 5 + len(self._tracers)):
            if int(forcing_digit(i)) > 0:
                f.append(get_forcing(i, int(forcing_digit(i))))
        return '\n'.join(f)




    @property
    def hgrid(self):
        return self._hgrid
    
    @property
    def tides(self):
        return self._tides

    @property
    def ntip(self):
        return len(self.tides.get_active_potential_constituents())

    @property
    def cutoff_depth(self):
        return float(self._cutoff_depth)

    @property
    def elevation(self):
        return self._elevation

    @property
    def velocity(self):
        return self._velocity
    
    @property
    def temperature(self):
        return self._temperature

    @property
    def salinity(self):
        return self._salinity


#! /usr/bin/env python
import unittest
from unittest.mock import patch
import sys
import pathlib
from pyschism.cmd import tidal_run


class TidalRunCmdTestCase(unittest.TestCase):

    def _test_tidal_run_mesh(self):
        parent = pathlib.Path(__name__).parent
        hgrid = parent.resolve() / "examples/example_1/hgrid.gr3"
        cmd = ["tidal_run"]
        cmd += [str(hgrid.resolve())]
        cmd += ["2017-9-18T12:00"]
        cmd += ["2017-9-23T12:00"]
        cmd += ["--spinup-days=2"]
        print(" ".join(cmd))
        with patch.object(sys, 'argv', cmd):
            self.assertEqual(tidal_run.main(), 0)

    def test_init(self):
        with patch.object(tidal_run, "main", return_value=0):
            with patch.object(tidal_run, "__name__", "__main__"):
                with patch.object(tidal_run.sys, 'exit') as mock_exit:
                    tidal_run.init()
                    assert mock_exit.call_args[0][0] == 0


if __name__ == '__main__':
    unittest.main()



import numpy as np
from matplotlib import rcParams
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize


def get_topobathy_kwargs(values, vmin, vmax, colors=256):
    vmin = np.min(values) if vmin is None else vmin
    vmax = np.max(values) if vmax is None else vmax
    if vmax <= 0.:
        cmap = plt.cm.seismic
        col_val = 0.
        levels = np.linspace(vmin, vmax, colors)
    else:
        wet_count = int(np.floor(
            colors*(float((values < 0.).sum()) / float(values.size))))
        col_val = float(wet_count)/colors
        dry_count = colors - wet_count
        colors_undersea = plt.cm.bwr(np.linspace(1., 0., wet_count))
        colors_land = plt.cm.terrain(np.linspace(0.25, 1., dry_count))
        colors = np.vstack((colors_undersea, colors_land))
        cmap = LinearSegmentedColormap.from_list('cut_terrain', colors)
        wlevels = np.linspace(vmin, 0.0, wet_count, endpoint=False)
        dlevels = np.linspace(0.0, vmax, dry_count)
        levels = np.hstack((wlevels, dlevels))
    if vmax > 0:
        norm = FixPointNormalize(
            sealevel=0.0,
            vmax=vmax,
            vmin=vmin,
            col_val=col_val
            )
    else:
        norm = None
    return {'cmap': cmap,
            'norm': norm,
            'levels': levels,
            'col_val': col_val,
            # 'extend': 'both'
            }


def get_axes(axes, figsize=None, subplot=111):
    figsize = rcParams["figure.figsize"] if figsize is None else figsize
    if axes is None:
        fig = plt.figure(figsize=figsize)
        axes = fig.add_subplot(subplot)
    return axes


class FixPointNormalize(Normalize):
    """
    This class is used for plotting. The reason it is declared here is that
    it is used by more than one submodule. In the future, this class will be
    native part of matplotlib. This definiton will be removed once the native
    matplotlib definition becomes available.
    Inspired by https://stackoverflow.com/questions/20144529/shifted-colorbar-matplotlib
    Subclassing Normalize to obtain a colormap with a fixpoint
    somewhere in the middle of the colormap.
    This may be useful for a `terrain` map, to set the "sea level"
    to a color in the blue/turquise range.
    """
    def __init__(self, vmin=None, vmax=None, sealevel=0, col_val=0.5,
                 clip=False):
        # sealevel is the fix point of the colormap (in data units)
        self.sealevel = sealevel
        # col_val is the color value in the range [0,1] that should represent
        # the sealevel.
        self.col_val = col_val
        Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        x, y = [self.vmin, self.sealevel, self.vmax], [0, self.col_val, 1]
        if np.ma.is_masked(value)is False:
            value = np.ma.masked_invalid(value)
        return np.ma.masked_where(value.mask, np.interp(value, x, y))


def figure(f):
    def decorator(*argv, **kwargs):
        axes = get_axes(
            kwargs.get('axes', None),
            kwargs.get('figsize', None)
            )
        kwargs.update({'axes': axes})
        axes = f(*argv, **kwargs)
        if kwargs.get('show', False):
            axes.axis('scaled')
            plt.show()
        return axes
    return decorator



#! /usr/bin/env python
import tempfile
import pathlib
import unittest
from pyschism.mesh import Vgrid


class VgridTestCase(unittest.TestCase):

    def test_init(self):
        v = Vgrid()
        v.boilerplate_2D
        self.assertIsInstance(v, Vgrid)

    def test_write(self):
        tmpdir = tempfile.TemporaryDirectory()
        v = Vgrid()
        v.write(pathlib.Path(tmpdir.name) / 'vgrid.in')
        self.assertIsInstance(v, Vgrid)

    def test_write_overwrite_raises(self):
        tmpfile = tempfile.NamedTemporaryFile()
        v = Vgrid()
        self.assertRaises(
            Exception,
            v.write,
            pathlib.Path(tmpfile.name)
        )



from datetime import timedelta
from typing import List, Union
import uuid


from pyschism.server.base import ServerConfig


# class SlurmJobList:

#     def __init__(self):
#         self.jobs = []

#     def __get__(self, obj, val):
#         return self.jobs


# class SlurmQueue:
#     jobs = SlurmJobList()

#     def __call__(self):
#         for job in self.jobs:
#             if job["depends_on"]


#     def add(self, callable, job_id=None, depends_on=None):
#         self.jobs.append({
#             id(callable) if job_id is None else job_id: callable,
#             "depends_on": depends_on})


class SlurmConfig(ServerConfig):
    """Configuration object for SLURM-enabled servers"""

    def __init__(
            self,
            account: str = None,
            ntasks: int = None,
            partition: str = None,
            walltime: timedelta = None,
            filename: str = None,
            run_directory: str = None,
            run_name: str = None,
            mail_type: str = None,
            mail_user: str = None,
            log_filename: str = None,
            modules: List[str] = None,
            schism_binary: str = None,
            extra_commands: List[str] = None,
            launcher: str = None,
            nodes: int = None,
            symlink_outputs: str = None,
            mpi_launcher: str = None,
    ):
        """
        Instantiate a new Slurm shell script (`*.job`).

        :param account: Slurm account name
        :param ntasks: number of total tasks for Slurm to run
        :param run_name: Slurm run name
        :param partition: partition to run on
        :param walltime: time delta
        :param driver_script_filename: file path to the driver shell script
        :param run_directory: directory to run in
        :param mail_type: email type
        :param mail_user: email address
        :param log_filename: file path to output log file
        :param modules: list of file paths to modules to load
        :param path_prefix: file path to prepend to the PATH
        :param extra_commands: list of extra shell commands to insert into script
        :param launcher: command to start processes on target system (`srun`, `ibrun`, etc.)
        :param nodes: number of total nodes
        :param schism_binary: path to schism binary to use
        """
        self.account = account
        self.nproc = ntasks
        self.run_name = run_name
        self.partition = partition
        self.walltime = walltime
        self.filename = filename
        self.run_directory = run_directory
        self.mail_type = mail_type
        self.mail_user = mail_user
        self.log_filename = log_filename
        self.modules = modules
        self.schism_binary = schism_binary
        self.extra_commands = extra_commands
        self.launcher = launcher
        self.nodes = nodes
        self.symlink_outputs = symlink_outputs
        self.mpi_launcher = mpi_launcher

    def __str__(self):
        f = [
            self.MPI_LAUNCHER,
            self.SCHISM_BINARY,
            self.SYMLINK_OUTPUTS,
            self.SLURM_NTASKS,
            self.SYMLINK_OUTPUTS,
            self.SLURM_ACCOUNT,
            self.SLURM_RUN_NAME,
            self.SLURM_PARTITION,
            self.SLURM_WALLTIME,
            self.SLURM_RUN_DIRECTORY,
            self.SLURM_MAIL_TYPE,
            self.SLURM_MAIL_USER,
            self.SLURM_LOG_FILE,
            self.SLURM_JOB_FILE,
        ]

        # self.modules = modules
        # self.path_prefix = path_prefix
        # self.extra_commands = extra_commands
        # self.launcher = launcher
        # self.nodes = nodes

        # f.extend([
        #     f if ,
        #     f,
        #     f"SLURM_JOB_FILE:={self.log_filename}",
        #     self.slurm,
        #     ])
        # f.append(self.slurm)
        return "\n".join(f)

    @property
    def walltime(self):
        if isinstance(self.__walltime, timedelta):
            hours, remainder = divmod(self.__walltime, timedelta(hours=1))
            minutes, remainder = divmod(remainder, timedelta(minutes=1))
            seconds = round(remainder / timedelta(seconds=1))
            return f'{hours:02}:{minutes:02}:{seconds:02}'
        return self.__walltime

    @walltime.setter
    def walltime(self, walltime: Union[timedelta, None]):
        self.__walltime = walltime

    @property
    def filename(self):
        return self.__filename

    @filename.setter
    def filename(self, filename):
        if filename is None:
            filename = 'slurm.job'
        self.__filename = filename

    @property
    def run_name(self):
        return self.__run_name

    @run_name.setter
    def run_name(self, run_name):
        if run_name is None:
            run_name = uuid.uuid4().hex
        self.__run_name = run_name

    @property
    def run_directory(self):
        return self.__run_directory

    @run_directory.setter
    def run_directory(self, run_directory):
        if run_directory is None:
            run_directory = '.'
        self.__run_directory = run_directory

    @property
    def log_filename(self):
        return self.__log_filename

    @log_filename.setter
    def log_filename(self, log_filename):
        if log_filename is None:
            log_filename = r"slurm.log"
        self.__log_filename = log_filename

    @property
    def mpi_launcher(self):
        return self.__mpi_launcher

    @mpi_launcher.setter
    def mpi_launcher(self, mpi_launcher):
        self.__mpi_launcher = "srun" if mpi_launcher is None \
            else mpi_launcher

    @property
    def SLURM_NTASKS(self):
        f = "SLURM_NTASKS="
        if self.nproc is not None:
            f += f"{self.nproc}"
        return f

    @property
    def SLURM_ACCOUNT(self):
        f = "SLURM_ACCOUNT="
        if self.account is not None:
            f += f"{self.account}"
        return f

    @property
    def SLURM_RUN_NAME(self):
        f = "SLURM_RUN_NAME="
        if self.run_name is not None:
            f += f"{self.run_name}"
        return f

    @property
    def SLURM_PARTITION(self):
        f = "SLURM_PARTITION="
        if self.partition is not None:
            f += f"{self.partition}"
        return f

    @property
    def SLURM_WALLTIME(self):
        f = "SLURM_WALLTIME="
        if self.walltime is not None:
            f += f"{self.walltime}"
        return f

    @property
    def SLURM_RUN_DIRECTORY(self):
        f = "SLURM_RUN_DIRECTORY="
        if self.run_directory is not None:
            f += f"{self.run_directory}"
        return f

    @property
    def SLURM_MAIL_TYPE(self):
        f = "SLURM_MAIL_TYPE="
        if self.mail_type is not None:
            f += f"{self.mail_type}"
        return f

    @property
    def SLURM_MAIL_USER(self):
        f = "SLURM_MAIL_USER="
        if self.mail_user is not None:
            f += f"{self.mail_user}"
        return f

    @property
    def SLURM_LOG_FILE(self):
        f = "SLURM_LOG_FILE="
        if self.log_filename is not None:
            f += f"{self.log_filename}"
        return f

    @property
    def SLURM_JOB_FILE(self):
        f = "SLURM_JOB_FILE="
        if self.filename is not None:
            f += f"{self.filename}"
        return f

    @property
    def MPI_LAUNCHER(self):
        f = "MPI_LAUNCHER="
        if self.mpi_launcher is not None:
            f += f"{self.mpi_launcher}"
        return f



from datetime import timezone, timedelta
import warnings

import matplotlib.pyplot as plt
import numpy as np

from pyschism.outputs import StationsOutput
from pyschism.utils.coops import CoopsDataCollector
from pyschism.enums import StationOutputIndex, StationOutputVariables


class PlotOutputStations:

    def __init__(self, stations: StationsOutput, obs=True):
        self.__stations = stations
        self._obs = obs
        if self._obs:
            self._coops = CoopsDataCollector()

    def plot(self, variable, station_index=None):
        filenames = [path.name for path in self.stations._manifest]
        var_index = StationOutputIndex[
            StationOutputVariables(variable).name].value + 1

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            data = np.loadtxt(
                self.stations._manifest[filenames.index(
                    f'staout_{var_index}')])
            if len(w) != 0:
                raise AttributeError(f'Empty record for variable {variable}.')

        if self.stations.start_date is not None:
            start_date = self.stations.start_date
            dates = [start_date + timedelta(seconds=t) for t in data[:, 0]]
        else:
            dates = data[:, 0]

        if station_index is None:
            for i in range(data.shape[1] - 2):
                plt.figure(i)
                if self._obs:
                    station_id = self.stations.get_station_id(i)
                    if station_id is not None:
                        obs = self._coops.fetch(
                            station_id, variable, self.stations.start_date,
                            self.stations.rndays)
                        plt.plot(obs['datetime'], obs['values'])
                        plt.title(obs['name'])
                        # fig, ax = plt.subplots(figsize=(8, 3))
                        # import matplotlib.dates as mdates
                        # plt.plot_date(obs['datetime'], obs['values'], ls='solid', lw=1.5, aa=True,marker='None',color='g')
                        # ax = plt.gca()
                        # dayFmt = mdates.DateFormatter('%a-%b-%d')
                        # hrFmt = mdates.DateFormatter('%H:00')
                        # ax.xaxis.set_major_formatter(dayFmt)
                        # ax.xaxis.set_major_locator(mdates.DayLocator())
                        # ax.xaxis.set_minor_locator(mdates.HourLocator(byhour=[12]))
                        # ax.xaxis.set_minor_formatter(hrFmt)
                        # plt.grid(b=True, which='both', linestyle='dotted')
                        # plt.show()
                plt.plot(dates, data[:, i+1])

        else:
            plt.plot(data[:, 0], data[:, station_index])
        plt.show()
        # import pdb; pdb.set_trace()
        # with open(self.manifest[filenames.index(f'staout_{index}')]) as f:
        #     for line in f:
        #         print(len(line.split()))
        #         exit()

    @property
    def stations(self):
        return self.__stations



#! /usr/bin/env python
import pathlib
import tempfile
from pyschism.driver.param import Param
import unittest


PARAM = (pathlib.Path(__file__)
         / "../../../pyschism/driver/param.nml").resolve()


class ParamTestCase(unittest.TestCase):

    def setUp(self):
        self.param = Param()

    def test_read(self):
        self.assertIsInstance(Param.open(PARAM), Param)

    def test_write(self):
        p = Param()
        tmpdir = tempfile.TemporaryDirectory()
        self.assertIsNone(p.write(pathlib.Path(tmpdir.name) / 'param.nml'))

    def test_write_overwrite_False_raises(self):
        file = tempfile.NamedTemporaryFile()
        p = Param()
        self.assertRaises(OSError, p.write, file.name)

    def test_core_ipre(self):
        self.param.ipre = 0
        self.assertEqual(self.param.core.ipre, 0)

    def test_core_ibc(self):
        self.param.ibc = 0
        self.assertEqual(self.param.core.ibc, 0)

    def test_core_ibtp(self):
        self.param.ibtp = 1
        self.assertEqual(self.param.core.ibtp, 1)

    def test_core_rnday(self):
        self.param.core.rnday = 30
        self.assertEqual(self.param.core.rnday, 30)

    def test_core_dt(self):
        self.param.timestep = 100.0
        self.assertEqual(self.param.core.dt, 100.0)

    def test_core_msc2(self):
        self.param.msc2 = 24
        self.assertEqual(self.param.core.msc2, 24)

    def test_core_mdc2(self):
        self.param.mdc2 = 30
        self.assertEqual(self.param.core.mdc2, 30)

    def test_core_ntracer_gen(self):
        self.param.ntracer_gen = 2
        self.assertEqual(self.param.core.ntracer_gen, 2)

    def test_core_ntracer_age(self):
        self.param.ntracer_age = 4
        self.assertEqual(self.param.core.ntracer_age, 4)

    def test_core_sed_class(self):
        self.param.sed_class = 5
        self.assertEqual(self.param.core.sed_class, 5)

    def test_core_eco_class(self):
        self.param.eco_class = 27
        self.assertEqual(self.param.core.eco_class, 27)

    def test_core_nspool(self):
        self.param.nspool = 36
        self.assertEqual(self.param.core.nspool, 36)

    def test_core_ihfskip(self):
        self.param.ihfskip = 864
        self.assertEqual(self.param.core.ihfskip, 864)

    def test_opt_ipre2(self):
        self.param.opt.ipre2 = 0
        self.assertEqual(self.param.opt.ipre2, 0)

    def test_opt_start_year(self):
        self.param.opt.start_year = 2000
        self.assertEqual(self.param.opt.start_year, 2000)

    def test_opt_start_month(self):
        self.param.opt.start_month = 1
        self.assertEqual(self.param.opt.start_month, 1)

    def test_opt_start_day(self):
        self.param.opt.start_day = 1
        self.assertEqual(self.param.opt.start_day, 1)

    def test_opt_start_hour(self):
        self.param.opt.start_hour = 0
        self.assertEqual(self.param.opt.start_hour, 0)

    def test_opt_utc_start(self):
        self.param.opt.utc_start = 8
        self.assertEqual(self.param.opt.utc_start, 8)

    def test_opt_ics(self):
        self.param.opt.ics = 1
        self.assertEqual(self.param.opt.ics, 1)

    def test_opt_ihot(self):
        self.param.opt.ihot = 0
        self.assertEqual(self.param.opt.ihot, 0)

    def test_opt_ieos_type(self):
        self.param.opt.ieos_type = 0
        self.assertEqual(self.param.opt.ieos_type, 0)

    def test_opt_ieos_pres(self):
        self.param.opt.ieos_pres = 0
        self.assertEqual(self.param.opt.ieos_pres, 0)

    def test_opt_eos_a(self):
        self.param.opt.eos_a = -0.1
        self.assertEqual(self.param.opt.eos_a, -0.1)

    def test_opt_eos_b(self):
        self.param.opt.eos_b = 1001.0
        self.assertEqual(self.param.opt.eos_b, 1001.0)

    def test_opt_nramp(self):
        self.param.opt.nramp = 1
        self.assertEqual(self.param.opt.nramp, 1)

    def test_opt_dramp(self):
        self.param.opt.dramp = 1.0
        self.assertEqual(self.param.opt.dramp, 1.0)

    def test_opt_nrampbc(self):
        self.param.opt.nrampbc = 0
        self.assertEqual(self.param.opt.nrampbc, 0)

    def test_opt_drampbc(self):
        self.param.opt.drampbc = 1.0
        self.assertEqual(self.param.opt.drampbc, 1.0)

    def test_opt_iupwind_mom(self):
        self.param.opt.iupwind_mom = 0
        self.assertEqual(self.param.opt.iupwind_mom, 0)

    def test_opt_indvel(self):
        self.param.opt.indvel = 0
        self.assertEqual(self.param.opt.indvel, 0)

    def test_opt_ihorcon(self):
        self.param.opt.ihorcon = 0
        self.assertEqual(self.param.opt.ihorcon, 0)

    def test_opt_hvis_coef0(self):
        self.param.opt.hvis_coef0 = 0.025
        self.assertEqual(self.param.opt.hvis_coef0, 0.025)

    def test_opt_ishapiro(self):
        self.param.opt.ishapiro = 1
        self.assertEqual(self.param.opt.ishapiro, 1)

    def test_opt_shapiro0(self):
        self.param.opt.shapiro0 = 0.5
        self.assertEqual(self.param.opt.shapiro0, 0.5)

    def test_opt_niter_shap(self):
        self.param.opt.niter_shap = 1
        self.assertEqual(self.param.opt.niter_shap, 1)

    def test_opt_thetai(self):
        self.param.opt.thetai = 0.6
        self.assertEqual(self.param.opt.thetai, 0.6)

    def test_opt_icou_elfe_wwm(self):
        self.param.opt.icou_elfe_wwm = 0
        self.assertEqual(self.param.opt.icou_elfe_wwm, 0)

    def test_opt_nstep_wwm(self):
        self.param.opt.nstep_wwm = 1
        self.assertEqual(self.param.opt.nstep_wwm, 1)

    def test_opt_iwbl(self):
        self.param.opt.iwbl = 0
        self.assertEqual(self.param.opt.iwbl, 0)

    def test_opt_hmin_radstress(self):
        self.param.opt.hmin_radstress = 1.0
        self.assertEqual(self.param.opt.hmin_radstress, 1.0)

    def test_opt_nrampwafo(self):
        self.param.opt.nrampwafo = 0
        self.assertEqual(self.param.opt.nrampwafo, 0)

    def test_opt_drampwafo(self):
        self.param.opt.drampwafo = 1.0
        self.assertEqual(self.param.opt.drampwafo, 1.0)

    def test_opt_turbinj(self):
        self.param.opt.turbinj = 0.15
        self.assertEqual(self.param.opt.turbinj, 0.15)

    def test_opt_imm(self):
        self.param.opt.imm = 0
        self.assertEqual(self.param.opt.imm, 0)

    def test_opt_ibdef(self):
        self.param.opt.ibdef = 10
        self.assertEqual(self.param.opt.ibdef, 10)

    def test_opt_slam0(self):
        self.param.opt.slam0 = -124
        self.assertEqual(self.param.opt.slam0, -124)

    def test_opt_sfea0(self):
        self.param.opt.sfea0 = 45
        self.assertEqual(self.param.opt.sfea0, 45)

    def test_opt_iunder_deep(self):
        self.param.opt.iunder_deep = 0
        self.assertEqual(self.param.opt.iunder_deep, 0)

    def test_opt_h1_bcc(self):
        self.param.opt.h1_bcc = 50.0
        self.assertEqual(self.param.opt.h1_bcc, 50.0)

    def test_opt_h2_bcc(self):
        self.param.opt.h2_bcc = 100.0
        self.assertEqual(self.param.opt.h2_bcc, 100.0)

    def test_opt_hw_depth(self):
        self.param.opt.hw_depth = 1000000.0
        self.assertEqual(self.param.opt.hw_depth, 1000000.0)

    def test_opt_hw_ratio(self):
        self.param.opt.hw_ratio = 0.5
        self.assertEqual(self.param.opt.hw_ratio, 0.5)

    def test_opt_ihydraulics(self):
        self.param.opt.ihydraulics = 0
        self.assertEqual(self.param.opt.ihydraulics, 0)

    def test_opt_if_source(self):
        self.param.opt.if_source = 0
        self.assertEqual(self.param.opt.if_source, 0)

    def test_opt_nramp_ss(self):
        self.param.opt.nramp_ss = 1
        self.assertEqual(self.param.opt.nramp_ss, 1)

    def test_opt_dramp_ss(self):
        self.param.opt.dramp_ss = 2
        self.assertEqual(self.param.opt.dramp_ss, 2)

    def test_opt_ihdif(self):
        self.param.opt.ihdif = 0
        self.assertEqual(self.param.opt.ihdif, 0)

    def test_opt_nchi(self):
        self.param.opt.nchi = 0
        self.assertEqual(self.param.opt.nchi, 0)

    def test_opt_dzb_min(self):
        self.param.opt.dzb_min = 0.5
        self.assertEqual(self.param.opt.dzb_min, 0.5)

    def test_opt_dzb_decay(self):
        self.param.opt.dzb_decay = 0.0
        self.assertEqual(self.param.opt.dzb_decay, 0.0)

    def test_opt_hmin_man(self):
        self.param.opt.hmin_man = 1.0
        self.assertEqual(self.param.opt.hmin_man, 1.0)

    def test_opt_ncor(self):
        self.param.opt.ncor = 0
        self.assertEqual(self.param.opt.ncor, 0)

    def test_opt_rlatitude(self):
        self.param.opt.rlatitude = 46
        self.assertEqual(self.param.opt.rlatitude, 46)

    def test_opt_coricoef(self):
        self.param.opt.coricoef = 0
        self.assertEqual(self.param.opt.coricoef, 0)

    def test_opt_ic_elev(self):
        self.param.opt.ic_elev = 0
        self.assertEqual(self.param.opt.ic_elev, 0)

    def test_opt_nramp_elev(self):
        self.param.opt.nramp_elev = 0
        self.assertEqual(self.param.opt.nramp_elev, 0)

    def test_opt_inv_atm_bnd(self):
        self.param.opt.inv_atm_bnd = 0
        self.assertEqual(self.param.opt.inv_atm_bnd, 0)

    def test_opt_prmsl_ref(self):
        self.param.opt.prmsl_ref = 101325.0
        self.assertEqual(self.param.opt.prmsl_ref, 101325.0)

    def test_opt_(self):
        self.param.opt.flag_ic = [1, 1, 0, None, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(self.param.opt.flag_ic, [1, 1, 0, None, 0, 0, 0, 0, 0, 0, 0])

    def test_opt_gen_wsett(self):
        self.param.opt.gen_wsett = 0.0001
        self.assertEqual(self.param.opt.gen_wsett, 0.0001)

    def test_opt_ibcc_mean(self):
        self.param.opt.ibcc_mean = 0
        self.assertEqual(self.param.opt.ibcc_mean, 0)

    def test_opt_rmaxvel(self):
        self.param.opt.rmaxvel = 10.0
        self.assertEqual(self.param.opt.rmaxvel, 10.0)

    def test_opt_velmin_btrack(self):
        self.param.opt.velmin_btrack = 0.0001
        self.assertEqual(self.param.opt.velmin_btrack, 0.0001)

    def test_opt_btrack_nudge(self):
        self.param.opt.btrack_nudge = 0.009013
        self.assertEqual(self.param.opt.btrack_nudge, 0.009013)

    def test_opt_ibtrack_openbnd(self):
        self.param.opt.ibtrack_openbnd = 1
        self.assertEqual(self.param.opt.ibtrack_openbnd, 1)

    def test_opt_ihhat(self):
        self.param.opt.ihhat = 1
        self.assertEqual(self.param.opt.ihhat, 1)

    def test_opt_inunfl(self):
        self.param.opt.inunfl = 0
        self.assertEqual(self.param.opt.inunfl, 0)

    def test_opt_h0(self):
        self.param.opt.h0 = 0.01
        self.assertEqual(self.param.opt.h0, 0.01)

    def test_opt_shorewafo(self):
        self.param.opt.shorewafo = 0
        self.assertEqual(self.param.opt.shorewafo, 0)

    def test_opt_moitn0(self):
        self.param.opt.moitn0 = 50
        self.assertEqual(self.param.opt.moitn0, 50)

    def test_opt_mxitn0(self):
        self.param.opt.mxitn0 = 1500
        self.assertEqual(self.param.opt.mxitn0, 1500)

    def test_opt_rtol0(self):
        self.param.opt.rtol0 = 1e-12
        self.assertEqual(self.param.opt.rtol0, 1e-12)

    def test_opt_nadv(self):
        self.param.opt.nadv = 1
        self.assertEqual(self.param.opt.nadv, 1)

    def test_opt_dtb_max(self):
        self.param.opt.dtb_max = 30.0
        self.assertEqual(self.param.opt.dtb_max, 30.0)

    def test_opt_dtb_min(self):
        self.param.opt.dtb_min = 10.0
        self.assertEqual(self.param.opt.dtb_min, 10.0)

    def test_opt_inter_mom(self):
        self.param.opt.inter_mom = 0
        self.assertEqual(self.param.opt.inter_mom, 0)

    def test_opt_kr_co(self):
        self.param.opt.kr_co = 1
        self.assertEqual(self.param.opt.kr_co, 1)

    def test_opt_itr_met(self):
        self.param.opt.itr_met = 3
        self.assertEqual(self.param.opt.itr_met, 3)

    def test_opt_h_tvd(self):
        self.param.opt.h_tvd = 5.0
        self.assertEqual(self.param.opt.h_tvd, 5.0)

    def test_opt_eps1_tvd_imp(self):
        self.param.opt.eps1_tvd_imp = 0.0001
        self.assertEqual(self.param.opt.eps1_tvd_imp, 0.0001)

    def test_opt_eps2_tvd_imp(self):
        self.param.opt.eps2_tvd_imp = 1e-14
        self.assertEqual(self.param.opt.eps2_tvd_imp, 1e-14)

    def test_opt_ip_weno(self):
        self.param.opt.ip_weno = 2
        self.assertEqual(self.param.opt.ip_weno, 2)

    def test_opt_courant_weno(self):
        self.param.opt.courant_weno = 0.5
        self.assertEqual(self.param.opt.courant_weno, 0.5)

    def test_opt_nquad(self):
        self.param.opt.nquad = 2
        self.assertEqual(self.param.opt.nquad, 2)

    def test_opt_ntd_weno(self):
        self.param.opt.ntd_weno = 1
        self.assertEqual(self.param.opt.ntd_weno, 1)

    def test_opt_epsilon1(self):
        self.param.opt.epsilon1 = 0.001
        self.assertEqual(self.param.opt.epsilon1, 0.001)

    def test_opt_epsilon2(self):
        self.param.opt.epsilon2 = 1e-10
        self.assertEqual(self.param.opt.epsilon2, 1e-10)

    def test_opt_i_prtnftl_weno(self):
        self.param.opt.i_prtnftl_weno = 0
        self.assertEqual(self.param.opt.i_prtnftl_weno, 0)

    def test_opt_epsilon3(self):
        self.param.opt.epsilon3 = 1e-25
        self.assertEqual(self.param.opt.epsilon3, 1e-25)

    def test_opt_ielad_weno(self):
        self.param.opt.ielad_weno = 0
        self.assertEqual(self.param.opt.ielad_weno, 0)

    def test_opt_small_elad(self):
        self.param.opt.small_elad = 0.0001
        self.assertEqual(self.param.opt.small_elad, 0.0001)

    def test_opt_nws(self):
        self.param.opt.nws = 0
        self.assertEqual(self.param.opt.nws, 0)

    def test_opt_wtiminc(self):
        self.param.opt.wtiminc = 150.0
        self.assertEqual(self.param.opt.wtiminc, 150.0)

    def test_opt_nrampwind(self):
        self.param.opt.nrampwind = 1
        self.assertEqual(self.param.opt.nrampwind, 1)

    def test_opt_drampwind(self):
        self.param.opt.drampwind = 1.0
        self.assertEqual(self.param.opt.drampwind, 1.0)

    def test_opt_iwindoff(self):
        self.param.opt.iwindoff = 0
        self.assertEqual(self.param.opt.iwindoff, 0)

    def test_opt_iwind_form(self):
        self.param.opt.iwind_form = -1
        self.assertEqual(self.param.opt.iwind_form, -1)

    def test_opt_impose_net_flux(self):
        self.param.opt.impose_net_flux = 0
        self.assertEqual(self.param.opt.impose_net_flux, 0)

    def test_opt_ihconsv(self):
        self.param.opt.ihconsv = 0
        self.assertEqual(self.param.opt.ihconsv, 0)

    def test_opt_isconsv(self):
        self.param.opt.isconsv = 0
        self.assertEqual(self.param.opt.isconsv, 0)

    def test_opt_itur(self):
        self.param.opt.itur = 3
        self.assertEqual(self.param.opt.itur, 3)

    def test_opt_dfv0(self):
        self.param.opt.dfv0 = 0.01
        self.assertEqual(self.param.opt.dfv0, 0.01)

    def test_opt_dfh0(self):
        self.param.opt.dfh0 = 0.0001
        self.assertEqual(self.param.opt.dfh0, 0.0001)

    def test_opt_mid(self):
        self.param.opt.mid = 'KL'
        self.assertEqual(self.param.opt.mid, 'KL')

    def test_opt_stab(self):
        self.param.opt.stab = 'KC'
        self.assertEqual(self.param.opt.stab, 'KC')

    def test_opt_xlsc0(self):
        self.param.opt.xlsc0 = 0.1
        self.assertEqual(self.param.opt.xlsc0, 0.1)

    def test_opt_inu_elev(self):
        self.param.opt.inu_elev = 0
        self.assertEqual(self.param.opt.inu_elev, 0)

    def test_opt_inu_uv(self):
        self.param.opt.inu_uv = 0
        self.assertEqual(self.param.opt.inu_uv, 0)

    def test_opt_inu_tr(self):
        self.param.opt.inu_tr = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(self.param.opt.inu_tr, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def test_opt_vnh1(self):
        self.param.opt.vnh1 = 400
        self.assertEqual(self.param.opt.vnh1, 400)

    def test_opt_vnf1(self):
        self.param.opt.vnf1 = 0.0
        self.assertEqual(self.param.opt.vnf1, 0.0)

    def test_opt_vnh2(self):
        self.param.opt.vnh2 = 500
        self.assertEqual(self.param.opt.vnh2, 500)

    def test_opt_vnf2(self):
        self.param.opt.vnf2 = 0.0
        self.assertEqual(self.param.opt.vnf2, 0.0)

    def test_opt_step_nu_tr(self):
        self.param.opt.step_nu_tr = 86400.0
        self.assertEqual(self.param.opt.step_nu_tr, 86400.0)

    def test_opt_h_bcc1(self):
        self.param.opt.h_bcc1 = 100.0
        self.assertEqual(self.param.opt.h_bcc1, 100.0)

    def test_opt_s1_mxnbt(self):
        self.param.opt.s1_mxnbt = 0.5
        self.assertEqual(self.param.opt.s1_mxnbt, 0.5)

    def test_opt_s2_mxnbt(self):
        self.param.opt.s2_mxnbt = 3.5
        self.assertEqual(self.param.opt.s2_mxnbt, 3.5)

    def test_opt_iharind(self):
        self.param.opt.iharind = 0
        self.assertEqual(self.param.opt.iharind, 0)

    def test_opt_iflux(self):
        self.param.opt.iflux = 0
        self.assertEqual(self.param.opt.iflux, 0)

    def test_opt_izonal5(self):
        self.param.opt.izonal5 = 0
        self.assertEqual(self.param.opt.izonal5, 0)

    def test_opt_ibtrack_test(self):
        self.param.opt.ibtrack_test = 0
        self.assertEqual(self.param.opt.ibtrack_test, 0)

    def test_opt_irouse_test(self):
        self.param.opt.irouse_test = 0
        self.assertEqual(self.param.opt.irouse_test, 0)

    def test_opt_flag_fib(self):
        self.param.opt.flag_fib = 1
        self.assertEqual(self.param.opt.flag_fib, 1)

    def test_opt_slr_rate(self):
        self.param.opt.slr_rate = 120.0
        self.assertEqual(self.param.opt.slr_rate, 120.0)

    def test_opt_isav(self):
        self.param.opt.isav = 0
        self.assertEqual(self.param.opt.isav, 0)

    def test_opt_sav_cd(self):
        self.param.opt.sav_cd = 1.13
        self.assertEqual(self.param.opt.sav_cd, 1.13)

    def test_opt_nstep_ice(self):
        self.param.opt.nstep_ice = 1
        self.assertEqual(self.param.opt.nstep_ice, 1)

    def test_opt_level_age(self):
        self.param.opt.level_age = [9, -999]
        self.assertEqual(self.param.opt.level_age, [9, -999])

    def test_opt_rearth_pole(self):
        self.param.opt.rearth_pole = 6378206.4
        self.assertEqual(self.param.opt.rearth_pole, 6378206.4)

    def test_opt_rearth_eq(self):
        self.param.opt.rearth_eq = 6378206.4
        self.assertEqual(self.param.opt.rearth_eq, 6378206.4)

    def test_opt_shw(self):
        self.param.opt.shw = 4184.0
        self.assertEqual(self.param.opt.shw, 4184.0)

    def test_opt_rho0(self):
        self.param.opt.rho0 = 1000.0
        self.assertEqual(self.param.opt.rho0, 1000.0)

    def test_opt_vclose_surf_frac(self):
        self.param.opt.vclose_surf_frac = 1.0
        self.assertEqual(self.param.opt.vclose_surf_frac, 1.0)

    def test_opt_iadjust_mass_consv0(self):
        self.param.opt.iadjust_mass_consv0 = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(self.param.opt.iadjust_mass_consv0, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def test_schout_nhot(self):
        self.param.schout.nhot = 0
        self.assertEqual(self.param.schout.nhot, 0)

    def test_schout_nhot_write(self):
        self.param.schout.nhot_write = 8640
        self.assertEqual(self.param.schout.nhot_write, 8640)

    def test_schout_iout_sta(self):
        self.param.schout.iout_sta = 0
        self.assertEqual(self.param.schout.iout_sta, 0)

    def test_schout_nspool_sta(self):
        self.param.schout.nspool_sta = 10
        self.assertEqual(self.param.schout.nspool_sta, 10)

    def test_schout_iof_hydro(self):
        self.param.schout.iof_hydro = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(self.param.schout.iof_hydro, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def test_schout_iof_wwm(self):
        self.param.schout.iof_wwm = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(self.param.schout.iof_wwm, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def test_schout_iof_gen(self):
        self.param.schout.iof_gen = [0, 0]
        self.assertEqual(self.param.schout.iof_gen, [0, 0])

    def test_schout_iof_age(self):
        self.param.schout.iof_age = [0, 0]
        self.assertEqual(self.param.schout.iof_age, [0, 0])

    def test_schout_iof_sed(self):
        self.param.schout.iof_sed = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(self.param.schout.iof_sed, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def test_schout_iof_eco(self):
        self.param.schout.iof_eco = [0]
        self.assertEqual(self.param.schout.iof_eco, [0])

    def test_schout_iof_icm(self):
        self.param.schout.iof_icm = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(self.param.schout.iof_icm, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def test_schout_iof_cos(self):
        self.param.schout.iof_cos = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(self.param.schout.iof_cos, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def test_schout_iof_fib(self):
        self.param.schout.iof_fib = [0]
        self.assertEqual(self.param.schout.iof_fib, [0])

    def test_schout_iof_sed2d(self):
        self.param.schout.iof_sed2d = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(self.param.schout.iof_sed2d, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def test_schout_iof_marsh(self):
        self.param.schout.iof_marsh = [0]
        self.assertEqual(self.param.schout.iof_marsh, [0])

    def test_schout_iof_ice(self):
        self.param.schout.iof_ice = [0, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(self.param.schout.iof_ice, [0, 0, 0, 0, 0, 0, 0, 0])

    def test_schout_iof_ana(self):
        self.param.schout.iof_ana = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(self.param.schout.iof_ana, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def test_elev(self):
        self.param.elev = 1
        self.assertEqual(self.param.elev, 1)

    def test_air_pressure(self):
        self.param.air_pressure = 1
        self.assertEqual(self.param.air_pressure, 1)

    def test_air_temperature(self):
        self.param.air_temperature = 1
        self.assertEqual(self.param.air_temperature, 1)

    def test_specific_humidity(self):
        self.param.specific_humidity = 1
        self.assertEqual(self.param.specific_humidity, 1)

    def test_solar_radiation(self):
        self.param.solar_radiation = 1
        self.assertEqual(self.param.solar_radiation, 1)

    def test_sensible_flux(self):
        self.param.sensible_flux = 1
        self.assertEqual(self.param.sensible_flux, 1)

    def test_latent_heat(self):
        self.param.latent_heat = 1
        self.assertEqual(self.param.latent_heat, 1)

    def test_upward_longwave(self):
        self.param.upward_longwave = 1
        self.assertEqual(self.param.upward_longwave, 1)

    def test_downward_longwave(self):
        self.param.downward_longwave = 1
        self.assertEqual(self.param.downward_longwave, 1)

    def test_total_heat_flux(self):
        self.param.total_heat_flux = 1
        self.assertEqual(self.param.total_heat_flux, 1)

    def test_evaporation(self):
        self.param.evaporation = 1
        self.assertEqual(self.param.evaporation, 1)

    def test_precipitation(self):
        self.param.precipitation = 1
        self.assertEqual(self.param.precipitation, 1)

    def test_bottom_stress(self):
        self.param.bottom_stress = 1
        self.assertEqual(self.param.bottom_stress, 1)

    def test_wind_speed(self):
        self.param.wind_speed = 1
        self.assertEqual(self.param.wind_speed, 1)

    def test_wind_stress(self):
        self.param.wind_stress = 1
        self.assertEqual(self.param.wind_stress, 1)

    def test_dahv(self):
        self.param.dahv = 1
        self.assertEqual(self.param.dahv, 1)

    def test_vertical_velocity(self):
        self.param.vertical_velocity = 1
        self.assertEqual(self.param.vertical_velocity, 1)

    def test_temp(self):
        self.param.temp = 1
        self.assertEqual(self.param.temp, 1)

    def test_salt(self):
        self.param.salt = 1
        self.assertEqual(self.param.salt, 1)

    def test_water_density(self):
        self.param.water_density = 1
        self.assertEqual(self.param.water_density, 1)

    def test_diffusivity(self):
        self.param.diffusivity = 1
        self.assertEqual(self.param.diffusivity, 1)

    def test_viscosity(self):
        self.param.viscosity = 1
        self.assertEqual(self.param.viscosity, 1)

    def test_TKE(self):
        self.param.TKE = 1
        self.assertEqual(self.param.TKE, 1)

    def test_mixing_length(self):
        self.param.mixing_length = 1
        self.assertEqual(self.param.mixing_length, 1)

    def test_hvel(self):
        self.param.hvel = 1
        self.assertEqual(self.param.hvel, 1)

    def test_hvel_side(self):
        self.param.hvel_side = 1
        self.assertEqual(self.param.hvel_side, 1)

    def test_wvel_elem(self):
        self.param.wvel_elem = 1
        self.assertEqual(self.param.wvel_elem, 1)

    def test_temp_elem(self):
        self.param.temp_elem = 1
        self.assertEqual(self.param.temp_elem, 1)

    def test_salt_elem(self):
        self.param.salt_elem = 1
        self.assertEqual(self.param.salt_elem, 1)

    def test_pressure_gradient(self):
        self.param.pressure_gradient = 1
        self.assertEqual(self.param.pressure_gradient, 1)

    def test_WWM_1(self):
        self.param.WWM_1 = 1
        self.assertEqual(self.param.WWM_1, 1)

    def test_WWM_2(self):
        self.param.WWM_2 = 1
        self.assertEqual(self.param.WWM_2, 1)

    def test_WWM_3(self):
        self.param.WWM_3 = 1
        self.assertEqual(self.param.WWM_3, 1)

    def test_WWM_4(self):
        self.param.WWM_4 = 1
        self.assertEqual(self.param.WWM_4, 1)

    def test_WWM_5(self):
        self.param.WWM_5 = 1
        self.assertEqual(self.param.WWM_5, 1)

    def test_WWM_6(self):
        self.param.WWM_6 = 1
        self.assertEqual(self.param.WWM_6, 1)

    def test_WWM_9(self):
        self.param.WWM_9 = 1
        self.assertEqual(self.param.WWM_9, 1)

    def test_WWM_10(self):
        self.param.WWM_10 = 1
        self.assertEqual(self.param.WWM_10, 1)

    def test_WWM_11(self):
        self.param.WWM_11 = 1
        self.assertEqual(self.param.WWM_11, 1)

    def test_WWM_12(self):
        self.param.WWM_12 = 1
        self.assertEqual(self.param.WWM_12, 1)

    def test_WWM_13(self):
        self.param.WWM_13 = 1
        self.assertEqual(self.param.WWM_13, 1)

    def test_WWM_14(self):
        self.param.WWM_14 = 1
        self.assertEqual(self.param.WWM_14, 1)

    def test_WWM_15(self):
        self.param.WWM_15 = 1
        self.assertEqual(self.param.WWM_15, 1)

    def test_WWM_16(self):
        self.param.WWM_16 = 1
        self.assertEqual(self.param.WWM_16, 1)

    def test_WWM_17(self):
        self.param.WWM_17 = 1
        self.assertEqual(self.param.WWM_17, 1)

    def test_WWM_18(self):
        self.param.WWM_18 = 1
        self.assertEqual(self.param.WWM_18, 1)

    def test_WWM_19(self):
        self.param.WWM_19 = 1
        self.assertEqual(self.param.WWM_19, 1)

    def test_WWM_20(self):
        self.param.WWM_20 = 1
        self.assertEqual(self.param.WWM_20, 1)

    def test_WWM_21(self):
        self.param.WWM_21 = 1
        self.assertEqual(self.param.WWM_21, 1)

    def test_WWM_22(self):
        self.param.WWM_22 = 1
        self.assertEqual(self.param.WWM_22, 1)

    def test_WWM_23(self):
        self.param.WWM_23 = 1
        self.assertEqual(self.param.WWM_23, 1)

    def test_WWM_24(self):
        self.param.WWM_24 = 1
        self.assertEqual(self.param.WWM_24, 1)

    def test_WWM_25(self):
        self.param.WWM_25 = 1
        self.assertEqual(self.param.WWM_25, 1)

    def test_WWM_26(self):
        self.param.WWM_26 = 1
        self.assertEqual(self.param.WWM_26, 1)

    def test_WWM_27(self):
        self.param.WWM_27 = 1
        self.assertEqual(self.param.WWM_27, 1)

    def test_WWM_28(self):
        self.param.WWM_28 = 1
        self.assertEqual(self.param.WWM_28, 1)

    def test_WWM_energy_dir(self):
        self.param.WWM_energy_dir = 1
        self.assertEqual(self.param.WWM_energy_dir, 1)

    def test_wave_force(self):
        self.param.wave_force = 1
        self.assertEqual(self.param.wave_force, 1)

    def test_GEN_1(self):
        self.param.GEN_1 = 1
        self.assertEqual(self.param.GEN_1, 1)

    def test_GEN_2(self):
        self.param.GEN_2 = 1
        self.assertEqual(self.param.GEN_2, 1)

    def test_AGE_1(self):
        self.param.AGE_1 = 1
        self.assertEqual(self.param.AGE_1, 1)

    def test_AGE_2(self):
        self.param.AGE_2 = 1
        self.assertEqual(self.param.AGE_2, 1)

    def test_SED_depth_change(self):
        self.param.SED_depth_change = 1
        self.assertEqual(self.param.SED_depth_change, 1)

    def test_SED_D50(self):
        self.param.SED_D50 = 1
        self.assertEqual(self.param.SED_D50, 1)

    def test_SED_bed_stress(self):
        self.param.SED_bed_stress = 1
        self.assertEqual(self.param.SED_bed_stress, 1)

    def test_SED_bed_roughness(self):
        self.param.SED_bed_roughness = 1
        self.assertEqual(self.param.SED_bed_roughness, 1)

    def test_SED_TSC(self):
        self.param.SED_TSC = 1
        self.assertEqual(self.param.SED_TSC, 1)

    def test_bed_thickness(self):
        self.param.bed_thickness = 1
        self.assertEqual(self.param.bed_thickness, 1)

    def test_bed_age(self):
        self.param.bed_age = 1
        self.assertEqual(self.param.bed_age, 1)

    def test_z0st(self):
        self.param.z0st = 1
        self.assertEqual(self.param.z0st, 1)

    def test_z0cr(self):
        self.param.z0cr = 1
        self.assertEqual(self.param.z0cr, 1)

    def test_z0sw(self):
        self.param.z0sw = 1
        self.assertEqual(self.param.z0sw, 1)

    def test_z0wr(self):
        self.param.z0wr = 1
        self.assertEqual(self.param.z0wr, 1)

    def test_SED3D_1(self):
        self.param.SED3D_1 = 1
        self.assertEqual(self.param.SED3D_1, 1)

    def test_SED_bdld_1(self):
        self.param.SED_bdld_1 = 1
        self.assertEqual(self.param.SED_bdld_1, 1)

    def test_SED_bedfrac_1(self):
        self.param.SED_bedfrac_1 = 1
        self.assertEqual(self.param.SED_bedfrac_1, 1)

    def test_SED3D_2(self):
        self.param.SED3D_2 = 1
        self.assertEqual(self.param.SED3D_2, 1)

    def test_SED_bdld_2(self):
        self.param.SED_bdld_2 = 1
        self.assertEqual(self.param.SED_bdld_2, 1)

    def test_SED_bedfrac_3(self):
        self.param.SED_bedfrac_3 = 1
        self.assertEqual(self.param.SED_bedfrac_3, 1)

    def test_ECO_1(self):
        self.param.ECO_1 = 1
        self.assertEqual(self.param.ECO_1, 1)

    def test_ICM_Chl(self):
        self.param.ICM_Chl = 1
        self.assertEqual(self.param.ICM_Chl, 1)

    def test_ICM_pH(self):
        self.param.ICM_pH = 1
        self.assertEqual(self.param.ICM_pH, 1)

    def test_ICM_PrmPrdt(self):
        self.param.ICM_PrmPrdt = 1
        self.assertEqual(self.param.ICM_PrmPrdt, 1)

    def test_ICM_DIN(self):
        self.param.ICM_DIN = 1
        self.assertEqual(self.param.ICM_DIN, 1)

    def test_ICM_PON(self):
        self.param.ICM_PON = 1
        self.assertEqual(self.param.ICM_PON, 1)

    def test_ICM_SED_BENDOC(self):
        self.param.ICM_SED_BENDOC = 1
        self.assertEqual(self.param.ICM_SED_BENDOC, 1)

    def test_ICM_SED_BENNH4(self):
        self.param.ICM_SED_BENNH4 = 1
        self.assertEqual(self.param.ICM_SED_BENNH4, 1)

    def test_ICM_SED_BENNO3(self):
        self.param.ICM_SED_BENNO3 = 1
        self.assertEqual(self.param.ICM_SED_BENNO3, 1)

    def test_ICM_SED_BENPO4(self):
        self.param.ICM_SED_BENPO4 = 1
        self.assertEqual(self.param.ICM_SED_BENPO4, 1)

    def test_ICM_SED_BENCOD(self):
        self.param.ICM_SED_BENCOD = 1
        self.assertEqual(self.param.ICM_SED_BENCOD, 1)

    def test_ICM_SED_BENDO(self):
        self.param.ICM_SED_BENDO = 1
        self.assertEqual(self.param.ICM_SED_BENDO, 1)

    def test_ICM_SED_BENSA(self):
        self.param.ICM_SED_BENSA = 1
        self.assertEqual(self.param.ICM_SED_BENSA, 1)

    def test_ICM_lfsav(self):
        self.param.ICM_lfsav = 1
        self.assertEqual(self.param.ICM_lfsav, 1)

    def test_ICM_stsav(self):
        self.param.ICM_stsav = 1
        self.assertEqual(self.param.ICM_stsav, 1)

    def test_ICM_rtsav(self):
        self.param.ICM_rtsav = 1
        self.assertEqual(self.param.ICM_rtsav, 1)

    def test_ICM_tlfsav(self):
        self.param.ICM_tlfsav = 1
        self.assertEqual(self.param.ICM_tlfsav, 1)

    def test_ICM_tstsav(self):
        self.param.ICM_tstsav = 1
        self.assertEqual(self.param.ICM_tstsav, 1)

    def test_ICM_trtsav(self):
        self.param.ICM_trtsav = 1
        self.assertEqual(self.param.ICM_trtsav, 1)

    def test_ICM_hcansav(self):
        self.param.ICM_hcansav = 1
        self.assertEqual(self.param.ICM_hcansav, 1)

    def test_ICM_CNH4(self):
        self.param.ICM_CNH4 = 1
        self.assertEqual(self.param.ICM_CNH4, 1)

    def test_ICM_CNH3(self):
        self.param.ICM_CNH3 = 1
        self.assertEqual(self.param.ICM_CNH3, 1)

    def test_ICM_CPIP(self):
        self.param.ICM_CPIP = 1
        self.assertEqual(self.param.ICM_CPIP, 1)

    def test_ICM_CPOS(self):
        self.param.ICM_CPOS = 1
        self.assertEqual(self.param.ICM_CPOS, 1)

    def test_ICM_CCH4(self):
        self.param.ICM_CCH4 = 1
        self.assertEqual(self.param.ICM_CCH4, 1)

    def test_ICM_CSO4(self):
        self.param.ICM_CSO4 = 1
        self.assertEqual(self.param.ICM_CSO4, 1)

    def test_ICM_CH2S(self):
        self.param.ICM_CH2S = 1
        self.assertEqual(self.param.ICM_CH2S, 1)

    def test_ICM_SEDPON1(self):
        self.param.ICM_SEDPON1 = 1
        self.assertEqual(self.param.ICM_SEDPON1, 1)

    def test_ICM_SEDPON2(self):
        self.param.ICM_SEDPON2 = 1
        self.assertEqual(self.param.ICM_SEDPON2, 1)

    def test_ICM_SEDPON3(self):
        self.param.ICM_SEDPON3 = 1
        self.assertEqual(self.param.ICM_SEDPON3, 1)

    def test_ICM_SEDPOP1(self):
        self.param.ICM_SEDPOP1 = 1
        self.assertEqual(self.param.ICM_SEDPOP1, 1)

    def test_ICM_SEDPOP2(self):
        self.param.ICM_SEDPOP2 = 1
        self.assertEqual(self.param.ICM_SEDPOP2, 1)

    def test_ICM_SEDPOP3(self):
        self.param.ICM_SEDPOP3 = 1
        self.assertEqual(self.param.ICM_SEDPOP3, 1)

    def test_ICM_SEDPOC1(self):
        self.param.ICM_SEDPOC1 = 1
        self.assertEqual(self.param.ICM_SEDPOC1, 1)

    def test_ICM_SEDPOC2(self):
        self.param.ICM_SEDPOC2 = 1
        self.assertEqual(self.param.ICM_SEDPOC2, 1)

    def test_ICM_SEDPOC3(self):
        self.param.ICM_SEDPOC3 = 1
        self.assertEqual(self.param.ICM_SEDPOC3, 1)

    def test_ICM_EROH2S(self):
        self.param.ICM_EROH2S = 1
        self.assertEqual(self.param.ICM_EROH2S, 1)

    def test_ICM_EROLPOC(self):
        self.param.ICM_EROLPOC = 1
        self.assertEqual(self.param.ICM_EROLPOC, 1)

    def test_ICM_ERORPOC(self):
        self.param.ICM_ERORPOC = 1
        self.assertEqual(self.param.ICM_ERORPOC, 1)

    def test_ICM_DO_consumption(self):
        self.param.ICM_DO_consumption = 1
        self.assertEqual(self.param.ICM_DO_consumption, 1)

    def test_ICM_GP1(self):
        self.param.ICM_GP1 = 1
        self.assertEqual(self.param.ICM_GP1, 1)

    def test_ICM_GP2(self):
        self.param.ICM_GP2 = 1
        self.assertEqual(self.param.ICM_GP2, 1)

    def test_ICM_GP3(self):
        self.param.ICM_GP3 = 1
        self.assertEqual(self.param.ICM_GP3, 1)

    def test_ICM_1(self):
        self.param.ICM_1 = 1
        self.assertEqual(self.param.ICM_1, 1)

    def test_ICM_2(self):
        self.param.ICM_2 = 1
        self.assertEqual(self.param.ICM_2, 1)

    def test_ICM_3(self):
        self.param.ICM_3 = 1
        self.assertEqual(self.param.ICM_3, 1)

    def test_ICM_4(self):
        self.param.ICM_4 = 1
        self.assertEqual(self.param.ICM_4, 1)

    def test_ICM_5(self):
        self.param.ICM_5 = 1
        self.assertEqual(self.param.ICM_5, 1)

    def test_ICM_6(self):
        self.param.ICM_6 = 1
        self.assertEqual(self.param.ICM_6, 1)

    def test_ICM_7(self):
        self.param.ICM_7 = 1
        self.assertEqual(self.param.ICM_7, 1)

    def test_ICM_8(self):
        self.param.ICM_8 = 1
        self.assertEqual(self.param.ICM_8, 1)

    def test_ICM_9(self):
        self.param.ICM_9 = 1
        self.assertEqual(self.param.ICM_9, 1)

    def test_ICM_10(self):
        self.param.ICM_10 = 1
        self.assertEqual(self.param.ICM_10, 1)

    def test_ICM_11(self):
        self.param.ICM_11 = 1
        self.assertEqual(self.param.ICM_11, 1)

    def test_ICM_12(self):
        self.param.ICM_12 = 1
        self.assertEqual(self.param.ICM_12, 1)

    def test_ICM_13(self):
        self.param.ICM_13 = 1
        self.assertEqual(self.param.ICM_13, 1)

    def test_ICM_14(self):
        self.param.ICM_14 = 1
        self.assertEqual(self.param.ICM_14, 1)

    def test_ICM_15(self):
        self.param.ICM_15 = 1
        self.assertEqual(self.param.ICM_15, 1)

    def test_ICM_16(self):
        self.param.ICM_16 = 1
        self.assertEqual(self.param.ICM_16, 1)

    def test_ICM_17(self):
        self.param.ICM_17 = 1
        self.assertEqual(self.param.ICM_17, 1)

    def test_ICM_18(self):
        self.param.ICM_18 = 1
        self.assertEqual(self.param.ICM_18, 1)

    def test_ICM_19(self):
        self.param.ICM_19 = 1
        self.assertEqual(self.param.ICM_19, 1)

    def test_ICM_20(self):
        self.param.ICM_20 = 1
        self.assertEqual(self.param.ICM_20, 1)

    def test_ICM_21(self):
        self.param.ICM_21 = 1
        self.assertEqual(self.param.ICM_21, 1)

    def test_ICM_22(self):
        self.param.ICM_22 = 1
        self.assertEqual(self.param.ICM_22, 1)

    def test_ICM_23(self):
        self.param.ICM_23 = 1
        self.assertEqual(self.param.ICM_23, 1)

    def test_ICM_24(self):
        self.param.ICM_24 = 1
        self.assertEqual(self.param.ICM_24, 1)

    def test_ICM_25(self):
        self.param.ICM_25 = 1
        self.assertEqual(self.param.ICM_25, 1)

    def test_COS_1(self):
        self.param.COS_1 = 1
        self.assertEqual(self.param.COS_1, 1)

    def test_COS_2(self):
        self.param.COS_2 = 1
        self.assertEqual(self.param.COS_2, 1)

    def test_COS_3(self):
        self.param.COS_3 = 1
        self.assertEqual(self.param.COS_3, 1)

    def test_COS_4(self):
        self.param.COS_4 = 1
        self.assertEqual(self.param.COS_4, 1)

    def test_COS_5(self):
        self.param.COS_5 = 1
        self.assertEqual(self.param.COS_5, 1)

    def test_COS_6(self):
        self.param.COS_6 = 1
        self.assertEqual(self.param.COS_6, 1)

    def test_COS_7(self):
        self.param.COS_7 = 1
        self.assertEqual(self.param.COS_7, 1)

    def test_COS_8(self):
        self.param.COS_8 = 1
        self.assertEqual(self.param.COS_8, 1)

    def test_COS_9(self):
        self.param.COS_9 = 1
        self.assertEqual(self.param.COS_9, 1)

    def test_COS_10(self):
        self.param.COS_10 = 1
        self.assertEqual(self.param.COS_10, 1)

    def test_COS_11(self):
        self.param.COS_11 = 1
        self.assertEqual(self.param.COS_11, 1)

    def test_COS_12(self):
        self.param.COS_12 = 1
        self.assertEqual(self.param.COS_12, 1)

    def test_COS_13(self):
        self.param.COS_13 = 1
        self.assertEqual(self.param.COS_13, 1)

    def test_FIB_1(self):
        self.param.FIB_1 = 1
        self.assertEqual(self.param.FIB_1, 1)

    def test_SED2D_depth_change(self):
        self.param.SED2D_depth_change = 1
        self.assertEqual(self.param.SED2D_depth_change, 1)

    def test_SED2D_cflsed(self):
        self.param.SED2D_cflsed = 1
        self.assertEqual(self.param.SED2D_cflsed, 1)

    def test_SED2D_d50(self):
        self.param.SED2D_d50 = 1
        self.assertEqual(self.param.SED2D_d50, 1)

    def test_SED2D_total_transport(self):
        self.param.SED2D_total_transport = 1
        self.assertEqual(self.param.SED2D_total_transport, 1)

    def test_SED2D_susp_load(self):
        self.param.SED2D_susp_load = 1
        self.assertEqual(self.param.SED2D_susp_load, 1)

    def test_SED2D_bed_load(self):
        self.param.SED2D_bed_load = 1
        self.assertEqual(self.param.SED2D_bed_load, 1)

    def test_SED2D_average_transport(self):
        self.param.SED2D_average_transport = 1
        self.assertEqual(self.param.SED2D_average_transport, 1)

    def test_SED2D_bottom_slope(self):
        self.param.SED2D_bottom_slope = 1
        self.assertEqual(self.param.SED2D_bottom_slope, 1)

    def test_z0eq(self):
        self.param.z0eq = 1
        self.assertEqual(self.param.z0eq, 1)

    def test_z0cr2d(self):
        self.param.z0cr2d = 1
        self.assertEqual(self.param.z0cr2d, 1)

    def test_z0sw2d(self):
        self.param.z0sw2d = 1
        self.assertEqual(self.param.z0sw2d, 1)

    def test_z0wr2d(self):
        self.param.z0wr2d = 1
        self.assertEqual(self.param.z0wr2d, 1)

    def test_marsh_flag(self):
        self.param.marsh_flag = 1
        self.assertEqual(self.param.marsh_flag, 1)

    def test_ICE_velocity(self):
        self.param.ICE_velocity = 1
        self.assertEqual(self.param.ICE_velocity, 1)

    def test_ICE_strain_rate(self):
        self.param.ICE_strain_rate = 1
        self.assertEqual(self.param.ICE_strain_rate, 1)

    def test_ICE_net_heat_flux(self):
        self.param.ICE_net_heat_flux = 1
        self.assertEqual(self.param.ICE_net_heat_flux, 1)

    def test_ICE_fresh_water_flux(self):
        self.param.ICE_fresh_water_flux = 1
        self.assertEqual(self.param.ICE_fresh_water_flux, 1)

    def test_ICE_top_T(self):
        self.param.ICE_top_T = 1
        self.assertEqual(self.param.ICE_top_T, 1)

    def test_ICE_tracer_1(self):
        self.param.ICE_tracer_1 = 1
        self.assertEqual(self.param.ICE_tracer_1, 1)

    def test_ICE_tracer_2(self):
        self.param.ICE_tracer_2 = 1
        self.assertEqual(self.param.ICE_tracer_2, 1)

    def test_ICE_tracer_3(self):
        self.param.ICE_tracer_3 = 1
        self.assertEqual(self.param.ICE_tracer_3, 1)

    def test_ANA_air_pres_grad_x(self):
        self.param.ANA_air_pres_grad_x = 1
        self.assertEqual(self.param.ANA_air_pres_grad_x, 1)

    def test_ANA_air_pres_grad_y(self):
        self.param.ANA_air_pres_grad_y = 1
        self.assertEqual(self.param.ANA_air_pres_grad_y, 1)

    def test_ANA_tide_pot_grad_x(self):
        self.param.ANA_tide_pot_grad_x = 1
        self.assertEqual(self.param.ANA_tide_pot_grad_x, 1)

    def test_ANA_tide_pot_grad_y(self):
        self.param.ANA_tide_pot_grad_y = 1
        self.assertEqual(self.param.ANA_tide_pot_grad_y, 1)

    def test_ANA_hor_viscosity_x(self):
        self.param.ANA_hor_viscosity_x = 1
        self.assertEqual(self.param.ANA_hor_viscosity_x, 1)

    def test_ANA_hor_viscosity_y(self):
        self.param.ANA_hor_viscosity_y = 1
        self.assertEqual(self.param.ANA_hor_viscosity_y, 1)

    def test_ANA_bclinic_force_x(self):
        self.param.ANA_bclinic_force_x = 1
        self.assertEqual(self.param.ANA_bclinic_force_x, 1)

    def test_ANA_bclinic_force_y(self):
        self.param.ANA_bclinic_force_y = 1
        self.assertEqual(self.param.ANA_bclinic_force_y, 1)

    def test_ANA_vert_viscosity_x(self):
        self.param.ANA_vert_viscosity_x = 1
        self.assertEqual(self.param.ANA_vert_viscosity_x, 1)

    def test_ANA_vert_viscosity_y(self):
        self.param.ANA_vert_viscosity_y = 1
        self.assertEqual(self.param.ANA_vert_viscosity_y, 1)

    def test_ANA_mom_advection_x(self):
        self.param.ANA_mom_advection_x = 1
        self.assertEqual(self.param.ANA_mom_advection_x, 1)

    def test_ANA_mom_advection_y(self):
        self.param.ANA_mom_advection_y = 1
        self.assertEqual(self.param.ANA_mom_advection_y, 1)

    def test_ANA_Richardson(self):
        self.param.ANA_Richardson = 1
        self.assertEqual(self.param.ANA_Richardson, 1)

    def test_ANA_transport_min_dt_elem(self):
        self.param.ANA_transport_min_dt_elem = 1
        self.assertEqual(self.param.ANA_transport_min_dt_elem, 1)


if __name__ == '__main__':
    unittest.main()



import pathlib


def reader(path):
    sms2dm = dict()
    with open(pathlib.Path(path), 'r') as f:
        f.readline()
        while 1:
            line = f.readline().split()
            if len(line) == 0:
                break
            if line[0] in ['E3T', 'E4Q']:
                if line[0] not in sms2dm:
                    sms2dm[line[0]] = {}
                sms2dm[line[0]].update({
                    line[1]: line[2:]
                    })
            if line[0] == 'ND':
                if line[0] not in sms2dm:
                    sms2dm[line[0]] = {}
                sms2dm[line[0]].update({
                    line[1]: (
                        list(map(float, line[2:-1])), float(line[-1])
                        )
                    })
    return sms2dm


def writer(sms2dm, path, overwrite=False):
    path = pathlib.Path(path)
    if path.is_file() and not overwrite:
        msg = 'File exists, pass overwrite=True to allow overwrite.'
        raise Exception(msg)
    with open(path, 'w') as f:
        f.write(string(sms2dm))
    return 0  # for unittests


def string(sms2dm):
    f = graph(sms2dm)
    f += boundaries(sms2dm)
    return f


def graph(sms2dm):
    f = "MESH2D\n"
    # TODO: Make faster using np.array2string
    f += triangular_elements(sms2dm)
    f += quadrilateral_elements(sms2dm)
    f += nodes(sms2dm)
    return f


def nodes(sms2dm):
    assert all(int(id) > 0 for id in sms2dm['ND'])
    f = ''
    for id, (coords, value)in sms2dm['ND'].items():
        f += f"ND {int(id):d} "
        f += f"{coords[0]:<.16E} "
        f += f"{coords[1]:<.16E} "
        f += f"{value:<.16E}\n"
    return f


def boundaries(sms2dm):
    f = ''
    if 'boundaries' in sms2dm.keys():
        for ibtype, bnds in sms2dm['boundaries'].items():
            for id, bnd in bnds.items():
                f += nodestring(bnd['indexes'])
    return f


def geom_string(geom_type, sms2dm):
    assert geom_type in ['E3T', 'E4Q', 'E6T', 'E8Q', 'E9Q']
    assert all(int(id) > 0 for id in sms2dm[geom_type])
    f = ''
    for id, geom in sms2dm[geom_type].items():
        f += f"{geom_type} {id} "
        for j in range(len(geom)):
            f += f"{geom[j]} "
        f += "\n"
    return f


def nodestring(geom):
    f = "NS "
    for i in range(len(geom) - 1):
        f += f"{geom[i]} "
    f += f"-{geom[-1]}\n"
    return f


def nodestrings(geom):
    pass


def triangular_elements(geom):
    f = ''
    if geom is not None:
        f += geom_string("E3T", geom)
    return f


def quadrilateral_elements(geom):
    f = ''
    if geom is not None:
        f += geom_string("E4Q", geom)
    return f



import os
import pathlib

from typing import Union


class Vgrid:

    def __init__(
        self,
        vgrid: Union[str, os.PathLike] = None
    ):
        """Represents a SCHISM vertical grid.

        Args:
            vgrid (optional): A path to a file or None. If vgrid is None, it
            is assummed that the user wants to use 2D.

        WARNING: This class only support 2D at the moment and it will ignore
        any inputs. This is so it can be used as a placeholder that outputs at
        least a 2D grid for minimalistic model configuration.
        """
        pass

    @staticmethod
    def open(path):
        raise NotImplementedError('Vgrid.open()')

    def __str__(self):
        return """2 !ivcor
2 1 1.e6 !nvrt, kz (# of Z-levels); h_s (transition depth between S and Z)
Z levels
1  -1.e6
S levels
40. 1. 1.e-4  !h_c, theta_b, theta_f
   1    -1.
   2    0."""

    def write(self, path, overwrite=False):
        path = pathlib.Path(path)
        if path.is_file() and not overwrite:
            msg = 'File exists, pass overwrite=True to allow overwrite.'
            raise Exception(msg)

        with open(path, 'w') as f:
            f.write(str(self))

    def is_2D(self):
        return True

    def is_3D(self):
        return False



from functools import lru_cache
import logging
from typing import Union

import geopandas as gpd
from matplotlib.cm import ScalarMappable
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
from shapely.geometry import LineString

from pyschism.figures import figure, get_topobathy_kwargs
from pyschism.mesh.parsers import grd
from pyschism.mesh.base import Gr3  # , sort_edges, signed_polygon_area

_logger = logging.getLogger(__name__)


class HgridBoundaries:

    def __init__(self, hgrid: "Hgrid", boundaries: Union[dict, None]):
        ocean_boundaries = []
        land_boundaries = []
        interior_boundaries = []
        if boundaries is not None:
            for ibtype, bnds in boundaries.items():
                if ibtype is None:
                    for id, data in bnds.items():
                        indexes = list(map(hgrid.nodes.get_index_by_id,
                                       data['indexes']))
                        ocean_boundaries.append({
                            'id': id,
                            "index_id": data['indexes'],
                            "indexes": indexes,
                            'geometry': LineString(hgrid.vertices[indexes])
                            })

                elif str(ibtype).endswith('1'):
                    for id, data in bnds.items():
                        indexes = list(map(hgrid.nodes.get_index_by_id,
                                       data['indexes']))
                        interior_boundaries.append({
                            'id': id,
                            'ibtype': ibtype,
                            "index_id": data['indexes'],
                            "indexes": indexes,
                            'geometry': LineString(hgrid.vertices[indexes])
                            })
                else:
                    for id, data in bnds.items():
                        _indexes = np.array(data['indexes'])
                        if _indexes.ndim > 1:
                            # ndim > 1 implies we're dealing with an ADCIRC
                            # mesh that includes boundary pairs, such as weir
                            new_indexes = []
                            for i, line in enumerate(_indexes.T):
                                if i % 2 != 0:
                                    new_indexes.extend(np.flip(line))
                                else:
                                    new_indexes.extend(line)
                            _indexes = np.array(new_indexes).flatten()
                        else:
                            _indexes = _indexes.flatten()
                        indexes = list(map(hgrid.nodes.get_index_by_id,
                                       _indexes))

                        land_boundaries.append({
                            'id': id,
                            'ibtype': ibtype,
                            "index_id": data['indexes'],
                            "indexes": indexes,
                            'geometry': LineString(hgrid.vertices[indexes])
                            })

        self._ocean = gpd.GeoDataFrame(ocean_boundaries)
        self._land = gpd.GeoDataFrame(land_boundaries)
        self._interior = gpd.GeoDataFrame(interior_boundaries)
        self._hgrid = hgrid
        self._data = boundaries

    def ocean(self):
        return self._ocean

    def land(self):
        return self._land

    def interior(self):
        return self._interior

    @property
    def data(self):
        return self._data

    @lru_cache(maxsize=1)
    def __call__(self):
        data = []
        for bnd in self.ocean().itertuples():
            data.append({
                'id': bnd.id,
                'ibtype': None,
                "index_id": bnd.index_id,
                "indexes": bnd.indexes,
                'geometry': bnd.geometry})

        for bnd in self.land().itertuples():
            data.append({
                'id': bnd.id,
                'ibtype': bnd.ibtype,
                "index_id": bnd.index_id,
                "indexes": bnd.indexes,
                'geometry': bnd.geometry})

        for bnd in self.interior().itertuples():
            data.append({
                'id': bnd.id,
                'ibtype': bnd.ibtype,
                "index_id": bnd.index_id,
                "indexes": bnd.indexes,
                'geometry': bnd.geometry})

        return gpd.GeoDataFrame(data, crs=self._hgrid.crs)

    def __len__(self):
        return len(self())


class Hgrid(Gr3):
    """
    Class that represents the unstructured planar mesh used by SCHISM.
    """
    # _boundaries = BoundariesDescriptor()

    def __init__(self, *args, boundaries=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._boundaries = HgridBoundaries(self, boundaries)

    @staticmethod
    def open(path, crs=None):
        if str(path).endswith('.ll') and crs is None:
            crs = 'epsg:4326'
        _grd = grd.read(path, crs=crs)
        _grd['nodes'] = {id: (coords, -val) for id, (coords, val)
                         in _grd['nodes'].items()}
        return Hgrid(**_grd)

    def to_dict(self):
        _grd = super().to_dict()
        _grd.update({
            "nodes": {id: (coord, -val) for id, (coord, val)
                      in self.nodes().items()},
            "boundaries": self.boundaries.data})
        return _grd

    @figure
    def make_plot(
        self,
        axes=None,
        vmin=None,
        vmax=None,
        show=False,
        title=None,
        # figsize=rcParams["figure.figsize"],
        extent=None,
        cbar_label=None,
        **kwargs
    ):
        if vmin is None:
            vmin = np.min(self.values)
        if vmax is None:
            vmax = np.max(self.values)
        kwargs.update(**get_topobathy_kwargs(self.values, vmin, vmax))
        kwargs.pop('col_val')
        levels = kwargs.pop('levels')
        if vmin != vmax:
            self.tricontourf(
                axes=axes,
                levels=levels,
                vmin=vmin,
                vmax=vmax,
                **kwargs
            )
        else:
            self.tripcolor(axes=axes, **kwargs)
        self.quadface(axes=axes, **kwargs)
        axes.axis('scaled')
        if extent is not None:
            axes.axis(extent)
        if title is not None:
            axes.set_title(title)
        mappable = ScalarMappable(cmap=kwargs['cmap'])
        mappable.set_array([])
        mappable.set_clim(vmin, vmax)
        divider = make_axes_locatable(axes)
        cax = divider.append_axes("bottom", size="2%", pad=0.5)
        cbar = plt.colorbar(
            mappable,
            cax=cax,
            orientation='horizontal'
        )
        cbar.set_ticks([vmin, vmax])
        cbar.set_ticklabels([np.around(vmin, 2), np.around(vmax, 2)])
        if cbar_label is not None:
            cbar.set_label(cbar_label)
        return axes

    @property
    def boundaries(self):
        return self._boundaries

    @property
    def ocean_boundaries(self):
        return self.boundaries.ocean

    @property
    def land_boundaries(self):
        return self.boundaries.land

    @property
    def interior_boundaries(self):
        return self.boundaries.interior


#     def auto_generate(
#         self,
#         threshold=0.,
#         land_ibtype=0,
#         interior_ibtype=1,
#     ):
#         values = self._obj.values
#         if np.any(np.isnan(values)):
#             raise Exception("Mesh contains invalid values. Raster values must"
#                             "be interpolated to the mesh before generating "
#                             "boundaries.")

#         # drop current boundaries
#         self._obj.__dict__['boundaries'] = {}
#         self._obj.__dict__['boundaries'][None] = {}

#         # generate exterior boundaries
#         for ring in self._obj.outer_ring_collection.values():
#             # find boundary edges
#             edge_tag = np.full(ring.shape, 0)
#             edge_tag[np.where(values[ring[:, 0]] < threshold)[0], 0] = -1
#             edge_tag[np.where(values[ring[:, 1]] < threshold)[0], 1] = -1
#             edge_tag[np.where(values[ring[:, 0]] >= threshold)[0], 0] = 1
#             edge_tag[np.where(values[ring[:, 1]] >= threshold)[0], 1] = 1
#             # sort boundary edges
#             ocean_boundary = list()
#             land_boundary = list()
#             for i, (e0, e1) in enumerate(edge_tag):
#                 if np.any(np.asarray((e0, e1)) == -1):
#                     ocean_boundary.append(tuple(ring[i, :]))
#                 elif np.any(np.asarray((e0, e1)) == 1):
#                     land_boundary.append(tuple(ring[i, :]))
#             ocean_boundaries = sort_edges(ocean_boundary)
#             land_boundaries = sort_edges(land_boundary)
#             _bnd_id = len(self._obj.boundaries[None])
#             for bnd in ocean_boundaries:
#                 e0, e1 = [list(t) for t in zip(*bnd)]
#                 e0 = list(map(self._obj.get_vertex_id_by_index, e0))
#                 data = e0 + [self._obj.get_vertex_id_by_index(e1[-1])]
#                 self._set_boundary_data(self._obj.vertex_id,
#                                         None, _bnd_id, data)
#                 _bnd_id += 1
#             # add land boundaries
#             if land_ibtype not in self._obj.boundaries:
#                 self._add_boundary_type(land_ibtype)
#             _bnd_id = len(self._obj.boundaries[land_ibtype])
#             for bnd in land_boundaries:
#                 e0, e1 = [list(t) for t in zip(*bnd)]
#                 e0 = list(map(self._obj.get_vertex_id_by_index, e0))
#                 data = e0 + [self._obj.get_vertex_id_by_index(e1[-1])]
#                 self._set_boundary_data(self._obj.vertex_id, land_ibtype,
#                                         _bnd_id, data)
#                 _bnd_id += 1
#         # generate interior boundaries
#         _bnd_id = 0
#         _interior_boundaries = defaultdict()
#         for interiors in self._obj.inner_ring_collection.values():
#             for interior in interiors:
#                 e0, e1 = [list(t) for t in zip(*interior)]
#                 if signed_polygon_area(self._obj._vertices[e0, :]) < 0:
#                     e0 = list(reversed(e0))
#                     e1 = list(reversed(e1))
#                 e0 = list(map(self._obj.get_vertex_id_by_index, e0))
#                 e0.append(e0[0])
#                 _interior_boundaries[_bnd_id] = e0
#                 _bnd_id += 1
#         self._add_boundary_type(interior_ibtype)
#         for bnd_id, data in _interior_boundaries.items():
#             self._set_boundary_data(self._obj._vertex_id, interior_ibtype,
#                                     bnd_id, data)

#     @figure
#     def plot(
#         self,
#         ibtype,
#         id,
#         tags=True,
#         axes=None,
#         show=False,
#         figsize=None,
#         **kwargs
#     ):
#         boundary = list(map(self._obj.get_vertex_index_by_id,
#                             self._obj.boundaries[ibtype][id].indexes))
#         p = axes.plot(self._obj.x[boundary], self._obj.y[boundary], **kwargs)
#         if tags:
#             axes.text(
#                 self._obj.x[boundary[len(boundary)//2]],
#                 self._obj.y[boundary[len(boundary)//2]],
#                 f"ibtype={ibtype}\nid={id}",
#                 color=p[-1].get_color()
#                 )
#         return axes

#     @figure
#     def make_plot(
#         self,
#         axes=None,
#         show=False,
#         figsize=None,
#         **kwargs
#     ):
#         kwargs.update({'axes': axes})
#         for ibtype, bnds in self._obj.boundaries.items():
#             for id in bnds:
#                 axes = self.plot(ibtype, id, **kwargs)
#                 kwargs.update({'axes': axes})
#         return kwargs['axes']




from datetime import timedelta
from enum import Enum
import logging
import pathlib
from typing import Union

import f90nml  # type: ignore[import]

from pyschism.enums import Stratification

_logger = logging.getLogger(__name__)

PARAM_TEMPLATE = pathlib.Path(__file__).parent / 'param.nml.template'
PARAM_DEFAULTS = f90nml.read(PARAM_TEMPLATE)['core']


class IntIbcType(Enum):
    BAROCLINIC = 0
    BAROTROPIC = 1

    @classmethod
    def _missing_(self, name):
        raise ValueError(f'{name} is not a valid integer for ibc. '
                         'Valid integers are 0 or 1.')


class StrIbcType(Enum):
    BAROCLINIC = 'baroclinic'
    BAROTROPIC = 'barotropic'

    @classmethod
    def _missing_(self, name):
        raise ValueError(f'{name} is not a valid string for ibc. '
                         'Valid strings are "barotropic" or "baroclinic".')


class Ipre:

    def __set__(self, obj, ipre: int):
        if ipre not in [0, 1]:
            raise ValueError("Argument to ipre attribute must be 0 or 1")
        obj.__dict__['ipre'] = ipre

    def __get__(self, obj, val):
        return obj.__dict__.get('ipre', 0)


class Ibc:

    def __set__(self, obj, ibc: Union[Stratification, int, str]):
        if isinstance(ibc, int):
            ibc = Stratification(IntIbcType(ibc).value)

        if not isinstance(ibc, Stratification):
            raise TypeError('Argument to attribute ibc must be of type '
                            f"{Stratification} or an 0, 1 integer or a string "
                            "'barotropic', 'baroclinic', not type "
                            f"{type(ibc)}.")

        obj.__dict__['ibc'] = ibc

    def __get__(self, obj, val):
        return obj.__dict__['ibc'].value


class Ibtp:

    def __set__(self, obj, ibtp: int):

        if ibtp not in [0, 1]:
            raise TypeError('Argument to attribute ibtp must be 0 or 1, not '
                            f"{ibtp}.")
        if ibtp == 1 and obj.ibc == 0:
            raise ValueError("ibtp cannot be set to 1 because ibc is equal to "
                             'zero.')
        obj.__dict__['ibtp'] = ibtp

    def __get__(self, obj, val):
        return obj.__dict__.get('ibtp', 0)


class Rnday:

    def __set__(self, obj, rnday: Union[float, timedelta]):

        if isinstance(rnday, float):
            rnday = timedelta(days=rnday)

        if not isinstance(rnday, timedelta):
            raise TypeError('Argument to attribute rnday must be of type float'
                            f' (to represent days) or type {timedelta}.')

        obj.__dict__['rnday'] = rnday

    def __get__(self, obj, val):
        return obj.__dict__['rnday']


class Dt:

    def __set__(self, obj, dt: Union[int, float, timedelta]):

        if isinstance(dt, (int, float)):
            dt = timedelta(seconds=dt)

        if not isinstance(dt, timedelta):
            raise TypeError('Argument to attribute dt must be of type float'
                            f' (to represent days) or type {timedelta}.')

        obj.__dict__['dt'] = dt

    def __get__(self, obj, val):
        return obj.__dict__['dt']


class Nspool:

    def __set__(self, obj, nspool: Union[int, float, timedelta, None]):
        if nspool is None:
            nspool = int(round(obj.rnday / obj.dt))
        if isinstance(nspool, timedelta):
            nspool = int(round(nspool / obj.dt))
        if isinstance(nspool, float):
            nspool = int(round(timedelta(hours=nspool) / obj.dt))
        if isinstance(nspool, (int, float)):
            if nspool <= 0:
                raise ValueError("nspool must be positive.")
        obj.__dict__['nspool'] = int(nspool)

    def __get__(self, obj, val):
        return obj.__dict__['nspool']


class Ihfskip:

    def __set__(self, obj, ihfskip: Union[int, timedelta, None]):

        if not isinstance(ihfskip, (int, timedelta, type(None))):
            raise TypeError('Argument ihfskip must be int, timedelta or None.')

        if ihfskip is None:
            ihfskip = int(round(obj.rnday / obj.dt))

        if isinstance(ihfskip, timedelta):
            ihfskip = int(round(ihfskip / obj.dt))

        if not (ihfskip / obj.nspool).is_integer():
            raise ValueError("ihfskip/nspool must be an integer but got  "
                             f"{ihfskip}/{obj.nspool}={ihfskip/obj.nspool}")

        obj.__dict__['ihfskip'] = ihfskip

    def __get__(self, obj, val):
        return obj.__dict__['ihfskip']


class CoreMeta(type):

    def __new__(meta, name, bases, attrs):
        for key, value in PARAM_DEFAULTS.items():
            if key in attrs:
                continue
            else:
                attrs[key] = value
        return type(name, bases, attrs)


class CORE(metaclass=CoreMeta):
    """ Provides error checking implementation for CORE group """

    ipre = Ipre()
    ibc = Ibc()
    ibtp = Ibtp()
    rnday = Rnday()
    dt = Dt()
    nspool = Nspool()
    ihfskip = Ihfskip()

    def __init__(
            self,
            ibc: Union[int, str, Stratification],
            rnday: Union[int, float, timedelta],
            dt: Union[int, float, timedelta],
            nspool: Union[int, float, timedelta] = None,
            ihfskip: Union[int, timedelta] = None,
    ):
        _logger.info('Initializing CORE')
        self.ibc = ibc
        self.rnday = rnday
        self.dt = dt
        self.nspool = nspool
        self.ihfskip = ihfskip
        self._mandatory = ['ipre', 'ibc', 'ibtp', 'rnday', 'dt', 'nspool',
                           'ihfskip']

    def __str__(self):
        data = []
        for key, default in PARAM_DEFAULTS.items():
            current = getattr(self, key)
            if key == 'rnday':
                current = current.total_seconds() / (60.*60.*24)
            if key == 'dt':
                current = current.total_seconds()
            if key in self._mandatory:
                data.append(f"  {key}={str(current)}")
            elif default != current:
                data.append(f"  {key}={str(current)}")
        data = '\n'.join(data)
        return f"&CORE\n{data}\n/"



from collections import OrderedDict
from datetime import datetime, timedelta, timezone
from enum import Enum
import logging
from typing import Union, Callable

import numpy as np

from pyschism.forcing.tides import bctypes
from pyschism.forcing.tides.tpxo import TPXO
from pyschism.forcing.tides.hamtide import HAMTIDE

_logger = logging.getLogger(__name__)


class TidalDatabase(Enum):
    TPXO = TPXO
    HAMTIDE = HAMTIDE


class StrTidalDatabase(Enum):
    TPXO = 'TPXO'
    HAMTIDE = 'HAMTIDE'

    @classmethod
    def _missing_(self, name):
        raise ValueError(f"{name} is not a valid tidal database.")


class TidalConstituents(Enum):
    M2 = "M2"
    # etc...

    @classmethod
    def _missing_(self, name):
        raise ValueError(f"{name} is not a valid tidal constituent.")


class ActiveConstituents:

    def __get__(self, obj, val):
        active_constituents = obj.__dict__.get('active_constituents')
        if active_constituents is None:
            active_constituents = OrderedDict()
            obj.__dict__['active_constituents'] = active_constituents
        return active_constituents

    def __set__(self, obj, val):
        return obj.__dict__['active_constituents']


class Tides(bctypes.BoundaryCondition):

    _active_constituents = ActiveConstituents()

    def __init__(
            self,
            elevation: bool = True,
            velocity: bool = False,
            database: Union[str, TidalDatabase] = TidalDatabase.HAMTIDE,
    ):
        """Main class for requesting tidal boundary forcing for a SCHISM run.

        The user should add tidal requests to this class via it's methods.

        Args:
            elevation (optional): Apply elevation forcing to boundary, defaults
                to True.
            velocity (optional): Apply velocity forcing to boundary (currently
                disabled), defaults to False.
            database (optional): Tidal database to use in order to obtain
                boundary initial conditions, defaults to TidalDatabase.TPXO
        """
        _logger.info('Initializing tidal boundary conditions.')
        if velocity is True and database == 'tpxo':
            raise NotImplementedError(
                'Boundary velocities are temporarily disabled for TPXO.')
        super().__init__(
            iettype=bctypes.InitialElevationType.TIDAL
            if elevation is True else bctypes.InitialElevationType.NONE,
            ifltype=bctypes.InitialFlowType.TIDAL
            if velocity is True else bctypes.InitialFlowType.NONE)

        self.forcing_database = database

    def __iter__(self):
        for constituent in self.active_constituents:
            yield (constituent, self.get_tidal_constituent(constituent))

    def __len__(self):
        return len(self.active_constituents)

    def __call__(self, start_date: datetime, rnday: Union[int, datetime],
                 constituent: Union[str, TidalConstituents]):
        """Returns the tidal characteristic values for a given time period.
        """
        _logger.debug(
            f'Computing tidal factors for start_date={start_date} and '
            f'constituent {constituent}')
        return (self.get_tidal_species_type(constituent),
                self.get_tidal_potential_amplitude(constituent),  # TPK
                self.get_orbital_frequency(constituent),  # FF*
                self.get_nodal_factor(start_date, rnday, constituent),  # Amig*
                self.get_greenwich_factor(start_date, rnday, constituent))  # FACE* # noqa:E501

    def get_elevation(self, constituent, vertices):
        return self.forcing_database.get_elevation(constituent, vertices)

    def get_velocity(self, constituent, vertices):
        return self.forcing_database.get_velocity(constituent, vertices)

    def use_all(self, potential=True, forcing=True):
        for constituent in self.all_constituents:
            if constituent not in self.tidal_potential_amplitudes:
                potential = False
            self._active_constituents[constituent] = {
                'potential': potential,
                'forcing': forcing
            }

    def use_major(self, potential=True, forcing=True):
        for constituent in self.major_constituents:
            self.use_constituent(constituent, potential, forcing)

    def use_constituent(self, constituent, potential=True, forcing=True):
        if constituent not in self.constituents:
            raise ValueError(f"Constituent must be one of {self.constituents}")
        if constituent not in self.tidal_potential_amplitudes:
            potential = False
        self._active_constituents[constituent] = {
            "potential": potential,
            "forcing": forcing,
        }

    def drop_constituent(self, constituent):
        if constituent not in self.active_constituents:
            raise ValueError("Argument constituent must be one of "
                             f"{self.active_constituents}")
        self._active_constituents.pop(constituent)

    def get_active_constituents(self):
        return list(self.active_constituents.keys())

    def get_active_forcing_constituents(self):
        return [const for const, _ in self.active_constituents.items()
                if _['forcing'] is True]

    def get_active_potential_constituents(self):
        return [const for const, _ in self.active_constituents.items()
                if _['potential'] is True]

    def get_tidal_potential_amplitude(self, constituent):
        if constituent in self.tidal_potential_amplitudes:
            return self.tidal_potential_amplitudes[constituent]

    def get_tidal_species_type(self, constituent):
        if constituent in self.tidal_species_type:
            return self.tidal_species_type[constituent]

    def get_orbital_frequency(self, constituent):
        return self.orbital_frequencies[constituent]

    def get_initial_conditions(self, constituent, vertices):
        return self.initial_conditions(constituent, vertices)

    def set_Z0(self, Z0, **kwargs):
        self._active_constituents['Z0'] = {
            'potential': False,
            'forcing': True
        }
        self.Z0 = Z0

    def _manage_dates(f: Callable):
        def decorator(self, start_date, rnday, constituent):
            val = f(self, start_date, rnday, constituent)
            del(self.start_date_utc)
            del(self.end_date_utc)
            return val
        return decorator

    @_manage_dates
    def get_nodal_factor(self, start_date: datetime,
                         rnday: Union[float, timedelta],
                         constituent: str):
        if start_date.tzinfo is not None and \
                start_date.tzinfo.utcoffset(start_date) is not None:
            self.start_date_utc = start_date.astimezone(timezone(timedelta(0)))
        else:
            self.start_date_utc = start_date

        if not isinstance(rnday, timedelta):
            rnday = timedelta(days=rnday)

        self.end_date_utc = self.start_date_utc + rnday
        if constituent == "M2":
            return self.EQ78
        elif constituent == "S2":
            return 1.0
        elif constituent == "N2":
            return self.EQ78
        elif constituent == "K1":
            return self.EQ227
        elif constituent == "M4":
            return (self.EQ78)**2.
        elif constituent == "O1":
            return self.EQ75
        elif constituent == "M6":
            return (self.EQ78)**3.
        elif constituent == "MK3":
            return self.EQ78*self.EQ227
        elif constituent == "S4":
            return 1.0
        elif constituent == "MN4":
            return (self.EQ78)**2.
        elif constituent == "Nu2":
            return self.EQ78
        elif constituent == "S6":
            return 1.0
        elif constituent == "MU2":
            return self.EQ78
        elif constituent == "2N2":
            return self.EQ78
        elif constituent == "OO1":
            return self.EQ77
        elif constituent == "lambda2":
            return self.EQ78
        elif constituent == "S1":
            return 1.0
        elif constituent == "M1":
            return self.EQ207
        elif constituent == "J1":
            return self.EQ76
        elif constituent == "Mm":
            return self.EQ73
        elif constituent == "Ssa":
            return 1.0
        elif constituent == "Sa":
            return 1.0
        elif constituent == "Msf":
            return self.EQ78
        elif constituent == "Mf":
            return self.EQ74
        elif constituent == "RHO":
            return self.EQ75
        elif constituent == "Q1":
            return self.EQ75
        elif constituent == "T2":
            return 1.0
        elif constituent == "R2":
            return 1.0
        elif constituent == "2Q1":
            return self.EQ75
        elif constituent == "P1":
            return 1.0
        elif constituent == "2SM2":
            return self.EQ78
        elif constituent == "M3":
            return self.EQ149
        elif constituent == "L2":
            return self.EQ215
        elif constituent == "2MK3":
            return self.EQ227*self.EQ78**2
        elif constituent == "K2":
            return self.EQ235
        elif constituent == "M8":
            return self.EQ78**4
        elif constituent == "MS4":
            return self.EQ78
        if constituent == "Z0":
            return self.Z0
        else:
            msg = f'Unrecognized constituent {constituent}'
            raise TypeError(msg)

    def _normalize_to_360(f: Callable):
        def decorator(self, start_date, rnday, constituent):
            return f(self, start_date, rnday, constituent) % 360.
        return decorator

    @_manage_dates
    @_normalize_to_360
    def get_greenwich_factor(self, start_date: datetime,
                             rnday: Union[float, timedelta],
                             constituent: str):
        if start_date.tzinfo is not None and \
                start_date.tzinfo.utcoffset(start_date) is not None:
            self.start_date_utc = start_date.astimezone(timezone(timedelta(0)))
        else:
            self.start_date_utc = start_date

        if not isinstance(rnday, timedelta):
            rnday = timedelta(days=rnday)

        self.end_date_utc = self.start_date_utc + rnday
        if constituent == "M2":
            return 2.*(self.DT-self.DS+self.DH)+2.*(self.DXI-self.DNU)
        elif constituent == "S2":
            return 2.*self.DT
        elif constituent == "N2":
            return 2.*(self.DT+self.DH)-3.*self.DS+self.DP+2. \
                * (self.DXI-self.DNU)
        elif constituent == "K1":
            return self.DT+self.DH-90.-self.DNUP
        elif constituent == "M4":
            return 4.*(self.DT-self.DS+self.DH)+4.*(self.DXI-self.DNU)
        elif constituent == "O1":
            return self.DT-2.*self.DS+self.DH+90.+2.*self.DXI-self.DNU
        elif constituent == "M6":
            return 6.*(self.DT-self.DS+self.DH)+6.*(self.DXI-self.DNU)
        elif constituent == "MK3":
            return 3.*(self.DT+self.DH)-2.*self.DS-90.+2.*(self.DXI-self.DNU) \
                - self.DNUP
        elif constituent == "S4":
            return 4.*self.DT
        elif constituent == "MN4":
            return 4.*(self.DT+self.DH)-5.*self.DS+self.DP+4.\
                * (self.DXI-self.DNU)
        elif constituent == "Nu2":
            return 2.*self.DT-3.*self.DS+4.*self.DH-self.DP+2. \
                * (self.DXI-self.DNU)
        elif constituent == "S6":
            return 6.*self.DT
        elif constituent == "MU2":
            return 2.*(self.DT+2.*(self.DH-self.DS))+2.*(self.DXI-self.DNU)
        elif constituent == "2N2":
            return 2.*(self.DT-2.*self.DS+self.DH+self.DP)+2. \
                * (self.DXI-self.DNU)
        elif constituent == "OO1":
            return self.DT+2.*self.DS+self.DH-90.-2.*self.DXI-self.DNU
        elif constituent == "lambda2":
            return 2.*self.DT-self.DS+self.DP+180.+2.*(self.DXI-self.DNU)
        elif constituent == "S1":
            return self.DT
        elif constituent == "M1":
            return self.DT-self.DS+self.DH-90.+self.DXI-self.DNU+self.DQ
        elif constituent == "J1":
            return self.DT+self.DS+self.DH-self.DP-90.-self.DNU
        elif constituent == "Mm":
            return self.DS-self.DP
        elif constituent == "Ssa":
            return 2.*self.DH
        elif constituent == "Sa":
            return self.DH
        elif constituent == "Msf":
            return 2.*(self.DS-self.DH)
        elif constituent == "Mf":
            return 2.*self.DS-2.*self.DXI
        elif constituent == "RHO":
            return self.DT+3.*(self.DH-self.DS)-self.DP+90.+2.\
                * self.DXI-self.DNU
        elif constituent == "Q1":
            return self.DT-3.*self.DS+self.DH+self.DP+90.+2.*self.DXI-self.DNU
        elif constituent == "T2":
            return 2.*self.DT-self.DH+self.DP1
        elif constituent == "R2":
            return 2.*self.DT+self.DH-self.DP1+180.
        elif constituent == "2Q1":
            return self.DT-4.*self.DS+self.DH+2.*self.DP+90.+2.*self.DXI \
                - self.DNU
        elif constituent == "P1":
            return self.DT-self.DH+90.
        elif constituent == "2SM2":
            return 2.*(self.DT+self.DS-self.DH)+2.*(self.DNU-self.DXI)
        elif constituent == "M3":
            return 3.*(self.DT-self.DS+self.DH)+3.*(self.DXI-self.DNU)
        elif constituent == "L2":
            return 2.*(self.DT+self.DH)-self.DS-self.DP+180.+2.\
                * (self.DXI-self.DNU)-self.DR
        elif constituent == "2MK3":
            return 3.*(self.DT+self.DH)-4.*self.DS+90.+4.*(self.DXI-self.DNU) \
                + self.DNUP
        elif constituent == "K2":
            return 2.*(self.DT+self.DH)-2.*self.DNUP2
        elif constituent == "M8":
            return 8.*(self.DT-self.DS+self.DH)+8.*(self.DXI-self.DNU)
        elif constituent == "MS4":
            return 2.*(2.*self.DT-self.DS+self.DH)+2.*(self.DXI-self.DNU)
        elif constituent == "Z0":
            return 0
        else:
            msg = f'Unrecognized constituent {constituent}'
            raise TypeError(msg)

    def get_lunar_node(self):
        return (259.1560564 - 19.328185764 * self.DYR - .0529539336 * self.DDAY
                - .0022064139 * self.hour_middle)

    def get_lunar_perigee(self):
        return (334.3837214 + 40.66246584 * self.DYR + .111404016 * self.DDAY
                + .004641834 * self.hour_middle)

    def get_lunar_mean_longitude(self):
        return (277.0256206 + 129.38482032 * self.DYR + 13.176396768
                * self.DDAY + .549016532 * self.start_date_utc.hour)

    def get_solar_perigee(self):
        return (281.2208569 + .01717836 * self.DYR + .000047064 * self.DDAY
                + .000001961 * self.start_date_utc.hour)

    def get_solar_mean_longitude(self):
        return (280.1895014 - .238724988 * self.DYR + .9856473288 * self.DDAY
                + .0410686387 * self.start_date_utc.hour)

    @property
    def EQ73(self):
        """ """
        return (2./3.-np.sin(self.I)**2)/.5021

    @property
    def EQ74(self):
        """ """
        return np.sin(self.I)**2/.1578

    @property
    def EQ75(self):
        """ """
        return np.sin(self.I)*np.cos(self.I/2.)**2/.37988

    @property
    def EQ76(self):
        """ """
        return np.sin(2.*self.I)/.7214

    @property
    def EQ77(self):
        """ """
        return np.sin(self.I)*np.sin(self.I/2.)**2/.0164

    @property
    def EQ78(self):
        """ """
        return (np.cos(self.I/2)**4)/.91544

    @property
    def EQ149(self):
        """ """
        return np.cos(self.I/2.)**6/.8758

    @property
    def EQ197(self):
        """ """
        return np.sqrt(2.310+1.435*np.cos(2.*(self.P - self.XI)))

    @property
    def EQ207(self):
        """ """
        return self.EQ75*self.EQ197

    @property
    def EQ213(self):
        """ """
        return np.sqrt(1.-12.*np.tan(self.I/2.)**2*np.cos(2.*self.P)
                       + 36.*np.tan(self.I/2.)**4)

    @property
    def EQ215(self):
        """ """
        return self.EQ78*self.EQ213

    @property
    def EQ227(self):
        """ """
        return np.sqrt(.8965*np.sin(2.*self.I)**2+.6001*np.sin(2.*self.I)
                       * np.cos(self.NU)+.1006)

    @property
    def EQ235(self):
        """ """
        return .001+np.sqrt(19.0444*np.sin(self.I)**4+2.7702*np.sin(self.I)**2
                            * np.cos(2.*self.NU)+.0981)

    @property
    def active_constituents(self):
        return self._active_constituents.copy()

    major_constituents = ('Q1', 'O1', 'P1', 'K1', 'N2', 'M2', 'S2', 'K2')

    minor_constituents = ('Mm', 'Mf', 'M4', 'MN4', 'MS4', '2N2', 'S1')

    @property
    def all_constituents(self):
        if isinstance(self.forcing_database, HAMTIDE):
            return self.major_constituents
        elif isinstance(self.forcing_database, TPXO):
            return (*self.major_constituents, *self.minor_constituents)
        else:
            raise NotImplementedError(
                'Unhandled forcing database instance of type '
                f'{type(self.forcing_database)}')

    orbital_frequencies = {
        'M4':      0.0002810378050173,
        'M6':      0.0004215567080107,
        'MK3':     0.0002134400613513,
        'S4':      0.0002908882086657,
        'MN4':     0.0002783986019952,
        'S6':      0.0004363323129986,
        'M3':      0.0002107783537630,
        '2MK3':    0.0002081166466594,
        'M8':      0.0005620756090649,
        'MS4':     0.0002859630068415,
        'M2':      0.0001405189025086,
        'S2':      0.0001454441043329,
        'N2':      0.0001378796994865,
        'Nu2':     0.0001382329037065,
        'MU2':     0.0001355937006844,
        '2N2':     0.0001352404964644,
        'lambda2': 0.0001428049013108,
        'T2':      0.0001452450073529,
        'R2':      0.0001456432013128,
        '2SM2':    0.0001503693061571,
        'L2':      0.0001431581055307,
        'K2':      0.0001458423172006,
        'K1':      0.0000729211583579,
        'O1':      0.0000675977441508,
        'OO1':     0.0000782445730498,
        'S1':      0.0000727220521664,
        'M1':      0.0000702594512543,
        'J1':      0.0000755603613800,
        'RHO':     0.0000653117453487,
        'Q1':      0.0000649585411287,
        '2Q1':     0.0000623193381066,
        'P1':      0.0000725229459750,
        'Mm':      0.0000026392030221,
        'Ssa':     0.0000003982128677,
        'Sa':      0.0000001991061914,
        'Msf':     0.0000049252018242,
        'Mf':      0.0000053234146919,
        'Z0':      0}

    tidal_potential_amplitudes = {
        'M2': 0.242334,
        'S2': 0.112841,
        'N2': 0.046398,
        'K2': 0.030704,
        'K1': 0.141565,
        'O1': 0.100514,
        'P1': 0.046843,
        'Q1': 0.019256,
        'Z0': 0}

    tidal_species_type = {
        'M2': 2,
        'S2': 2,
        'N2': 2,
        'K2': 2,
        'K1': 1,
        'O1': 1,
        'P1': 1,
        'Q1': 1,
        'Z0': 0}

    @property
    def hour_middle(self):
        return self.start_date_utc.hour + (
            (self.end_date_utc - self.start_date_utc).total_seconds()
            / 3600 / 2)

    @property
    def I(self):  # noqa:E741,E743
        return np.arccos(.9136949-.0356926*np.cos(self.N))

    @property
    def N(self):
        return np.deg2rad(self.DN)

    @property
    def DN(self):
        return self.get_lunar_node()

    @property
    def DYR(self):
        return self.start_date_utc.year - 1900.

    @property
    def DDAY(self):
        return (self.start_date_utc.timetuple().tm_yday
                + int((self.start_date_utc.year-1901.)/4.) - 1)

    @property
    def NU(self):
        return np.arcsin(.0897056*np.sin(self.N)/np.sin(self.I))

    @property
    def DT(self):
        return (180.+self.start_date_utc.hour*(360./24))

    @property
    def DS(self):
        return self.get_lunar_mean_longitude()

    @property
    def DP(self):
        return self.get_lunar_perigee()

    @property
    def P(self):
        return np.deg2rad(self.DP)

    @property
    def DH(self):
        return self.get_solar_mean_longitude()

    @property
    def DP1(self):
        return self.get_solar_perigee()  # HR

    @property
    def DNU(self):
        return np.rad2deg(self.NU)

    @property
    def XI(self):
        return self.N-2.*np.arctan(.64412*np.tan(self.N/2)) - self.NU

    @property
    def DXI(self):
        return np.rad2deg(self.XI)

    @property
    def NUP(self):
        return np.arctan(np.sin(self.NU) / (
            np.cos(self.NU)+.334766/np.sin(2.*self.I)))

    @property
    def DNUP(self):
        return np.rad2deg(self.NUP)

    @property
    def DPC(self):
        return self.DP - self.DXI

    @property
    def PC(self):
        return np.deg2rad(self.DPC)

    @property
    def R(self):
        return (np.arctan(np.sin(2.*self.PC)
                          / ((1./6.)*(1./np.tan(.5*self.I))**2 - np.cos(2.*self.PC))))

    @property
    def DR(self):
        return np.rad2deg(self.R)

    @property
    def NUP2(self):
        return (np.arctan(np.sin(2.*self.NU)
                          / (np.cos(2.*self.NU)+.0726184 / np.sin(self.I)**2))/2.)

    @property
    def DNUP2(self):
        return np.rad2deg(self.NUP2)

    @property
    def Q(self):
        return np.arctan2((5.*np.cos(self.I)-1.)*np.sin(self.PC),
                          (7.*np.cos(self.I)+1.)*np.cos(self.PC))

    @property
    def DQ(self):
        return np.rad2deg(self.Q)

    @property
    def ntip(self):
        return len(self.get_active_potential_constituents())

    @property
    def cutoff_depth(self):
        if self.ntip == 0:
            return 0
        try:
            return self.__cutoff_depth
        except AttributeError:
            return 50.

    @cutoff_depth.setter
    def cutoff_depth(self, cutoff_depth):
        assert isinstance(cutoff_depth, (int, float))
        self.__cutoff_depth = cutoff_depth

    @property
    def nbfr(self):
        return len(self.get_active_forcing_constituents())

    @property
    def forcing_database(self):
        return self._forcing_database

    @forcing_database.setter
    def forcing_database(self, database: Union[str, TidalDatabase]):
        database = 'hamtide' if database is None else database
        if isinstance(database, str):
            database = TidalDatabase[
                StrTidalDatabase(database.upper()).name].value()
        if isinstance(database, TidalDatabase):
            database = database.value()
        self._forcing_database = database

    @property
    def constituents(self):
        return self.all_constituents

    @property
    def active_constituents(self):
        return self._active_constituents.copy()



#! /usr/bin/env python
import unittest
import pathlib
import tempfile
from pyschism.mesh.grd import reader, writer, euclidean_mesh


class GrdTestCase(unittest.TestCase):

    def setUp(self):
        self.nodes = {
            '1': ((0., 0.), -99999.),
            '2': ((.5, 0.), -99999.),
            '3': ((1., 0.), -99999.),
            '4': ((1., 1.), -99999.),
            '5': ((0., 1.), -99999.),
            '6': ((.5, 1.5), -99999.),
            '7': ((.33, .33), -99999.),
            '8': ((.66, .33), -99999.),
            '9': ((.5, .66), -99999.),
            '10': ((-1., 1.), -99999.),
            '11': ((-1., 0.), -99999.),
            }
        self.elements = {
            '1': ['5', '7', '9'],
            '2': ['1', '2', '7'],
            '3': ['2', '3', '8'],
            '4': ['8', '7', '2'],
            '5': ['3', '4', '8'],
            '6': ['4', '9', '8'],
            '7': ['4', '6', '5'],
            '8': ['5', '10', '11', '1'],
            '9': ['9', '4', '5'],
            '10': ['5', '1', '7']
            }

        self.boundaries = dict()

        self.boundaries[None] = {  # "open" boundaries
                0: {'indexes': ['10', '11', '1', '2']},
                1: {'indexes': ['2', '3', '4']}
        }

        self.boundaries[0] = {  # "land" boundaries
            0: {'indexes': ['4', '6']},
            1: {'indexes': ['6',  '5', '10']}
        }

        self.boundaries[1] = {  # "interior" boundary
            0: {'indexes': ['7', '8', '9', '7']}
        }

        self.grd = {
            'nodes': self.nodes,
            'elements': self.elements,
            'boundaries': self.boundaries,
            'description': 'grd_unittest'
        }

    def test_write_read(self):
        tmpdir = tempfile.TemporaryDirectory()
        tmpfile = pathlib.Path(tmpdir.name) / 'hgrid.grd'
        writer(self.grd, pathlib.Path(tmpfile))
        self.assertDictEqual(reader(pathlib.Path(tmpfile)), self.grd)

    def test_overwrite(self):
        tmpdir = tempfile.TemporaryDirectory()
        writer(self.grd, pathlib.Path(tmpdir.name) / 'hgrid.grd')
        self.assertRaises(
            Exception,
            writer,
            self.grd,
            pathlib.Path(tmpdir.name) / 'hgrid.grd'
            )

    def test_no_ocean_bnd(self):
        tmpdir = tempfile.TemporaryDirectory()
        tmpfile = pathlib.Path(tmpdir.name) / 'hgrid.grd'
        self.grd['boundaries'].pop(None)
        writer(self.grd, pathlib.Path(tmpfile))
        self.assertDictEqual(reader(pathlib.Path(tmpfile)), self.grd)

    def test_euclidean_mesh(self):
        self.assertIsInstance(euclidean_mesh(self.grd), dict)


if __name__ == '__main__':
    unittest.main()



from pyschism.forcing.tides import Tides
from pyschism.forcing.atmosphere import GlobalForecastSystem
from pyschism.forcing.hydrology.base import Hydrology

__all__ = [
    "Tides",
    "GlobalForecastSystem",
    "Hydrology"
]



from argparse import Namespace
from datetime import datetime, timedelta
import json
# import logging
import os
import pathlib
import shutil
import subprocess

import numpy as np  # type: ignore[import]
import pytz

from pyschism.domain import ModelDomain
from pyschism.driver import ModelDriver
from pyschism.forcing import Tides
from pyschism.mesh.vgrid import Vgrid
from pyschism.server import SlurmConfig

# from pyschism.logger import get_logger


class ProjectDirectory:

    def __get__(self, obj, val):
        project_directory = obj.__dict__.get('project_directory')
        if project_directory is None:
            project_directory = pathlib.Path(obj.args.project_directory)
            project_directory.mkdir(parents=True, exist_ok=True)
            obj.__dict__['project_directory'] = project_directory
        return project_directory


class ConfigFile:

    def __get__(self, obj, val):
        config_file = obj.__dict__.get('config_file')
        if config_file is None:
            config_file = obj.project_directory / 'config.json'
            obj.__dict__['config_file'] = config_file
        return config_file


class ColdstartDirectory:

    def __get__(self, obj, val):
        coldstart_directory = obj.__dict__.get('coldstart_directory')
        if coldstart_directory is None:
            timestamp = str(obj.target_datetime).replace(' ', 'T')
            coldstart_directory = obj.project_directory / \
                'coldstart' / f'{timestamp}'
            coldstart_directory.mkdir(
                parents=True, exist_ok=True)
            obj.__dict__['coldstart_directory'] = coldstart_directory
        return coldstart_directory


class TargetDatetime:

    def __get__(self, obj, val):
        target_datetime = obj.__dict__.get('target_datetime')
        if target_datetime is None:
            utcnow = pytz.timezone('UTC').localize(datetime.utcnow())
            # localnow = utcnow.astimezone(pytz.timezone(obj.args.timezone))
            nowcast_cycle = int(obj.forecast_interval * np.floor(
                utcnow.hour/obj.forecast_interval))  # % 24
            target_datetime = datetime(
                utcnow.year, utcnow.month, utcnow.day, nowcast_cycle,
                tzinfo=pytz.timezone(obj.args.timezone))
            obj.__dict__['target_datetime'] = target_datetime
        return target_datetime


class ColdstartDomain:

    def __get__(self, obj, val):
        model_domain = obj.__dict__.get('coldstart_domain')
        if model_domain is None:
            model_domain = ModelDomain.open(
                obj.hgrid_path,
                obj.fgrid_path,
                # TODO: Vgrid.open() raises NotImplementedError
                # obj.vgrid_path,
                hgrid_crs=obj.args.hgrid_crs,
                fgrid_crs=obj.args.fgrid_crs
            )
            if obj.tides is not None:
                model_domain.add_boundary_condition(obj.tides)
        return model_domain


class ColdstartDriver:

    def __get__(self, obj, val):
        coldstart = obj.__dict__.get('coldstart_driver')
        if coldstart is None:
            start_date = obj.target_datetime - timedelta(
                days=obj.args.spinup_days)
            rnday = obj.target_datetime - start_date
            coldstart = ModelDriver(
                    model_domain=obj.coldstart_domain,
                    dt=obj.args.timestep,
                    rnday=rnday,
                    dramp=(2./3.)*rnday,
                    start_date=start_date,
                    nhot_write=True,
                    # ibc=obj.ibc,
                    # drampbc=obj.drampbc,
                    server_config=obj.server_config,
                    )
            obj.__dict__['coldstart_driver'] = coldstart
        return coldstart


class ForecastInterval:

    def __get__(self, obj, val):
        forecast_interval = obj.__dict__.get('forecast_interval')
        if forecast_interval is None:
            if obj.args.forecast_interval != 24:
                raise NotImplementedError(
                    'Forecast interval only at 24 hours for now.')
            obj.__dict__['forecast_interval'] = obj.args.forecast_interval
        return obj.__dict__['forecast_interval']


class HgridPath:

    def __get__(self, obj, val):
        hgrid_path = obj.__dict__.get('hgrid_path')
        if hgrid_path is None:
            hgrid_path = obj.static_files_directory / 'hgrid.gr3'
            if hgrid_path.is_file() and obj.args.overwrite is True:
                hgrid_path.unlink()
            if not hgrid_path.is_file():
                shutil.copy2(obj.args.hgrid, hgrid_path,
                             follow_symlinks=True)
            obj.__dict__['hgrid_path'] = hgrid_path
        return hgrid_path


class VgridPath:

    def __get__(self, obj, val):
        vgrid_path = obj.__dict__.get('vgrid_path')
        if vgrid_path is None:
            vgrid_path = obj.static_files_directory / 'vgrid.in'
            if vgrid_path.is_file():
                if obj.args.overwrite is True:
                    vgrid_path.unlink()
            if obj.args.vgrid is None:
                if not vgrid_path.exists():
                    Vgrid().write(vgrid_path)
            else:
                shutil.copy2(
                    obj.args.vgrid, vgrid_path, follow_symlinks=True)
            obj.__dict__['vgrid_path'] = vgrid_path
        return vgrid_path


class FgridPath:

    def __get__(self, obj, val):
        fgrid_path = obj.__dict__.get('fgrid_path')
        if fgrid_path is None:
            if obj.args.fgrid_type == 'auto':
                fgrid = pathlib.Path(obj.args.fgrid)
                fgrid_path = obj.static_files_directory / fgrid.name
            else:
                fgrid_path = obj.static_files_directory / \
                    (obj.args.fgrid_type + '.gr3')
            if obj.args.overwrite is True:
                if fgrid_path.is_file():
                    os.remove(fgrid_path)
            shutil.copy2(obj.args.fgrid, fgrid_path, follow_symlinks=True)
            obj.__dict__['fgrid_path'] = fgrid_path
        return fgrid_path


class StaticFilesDirectory:

    def __get__(self, obj, val):
        static_files = obj.__dict__.get('static_files')
        if static_files is None:
            static_files = obj.project_directory / 'static'
            static_files.mkdir(exist_ok=True)
            obj.__dict__['static_files'] = static_files
        return static_files


class TidesDescriptor:

    def __get__(self, obj, val):
        tides = obj.__dict__.get('tides')
        if tides is None:
            if not obj.args.all_constituents \
                    and not obj.args.major_constituents \
                    and not obj.args.constituents:
                return
            else:
                tides = Tides(velocity=obj.args.bnd_vel)
                if obj.args.all_constituents:
                    tides.use_all()
                if obj.args.major_constituents:
                    tides.use_major()
                if obj.args.constituents:
                    for constituent in obj.args.constituents:
                        tides.use_constituent(constituent)
            obj.__dict__['tides'] = tides
        return tides


class ServerConfigDescriptor:

    def __get__(self, obj, val):
        server_config = obj.__dict__.get("server_config")
        if server_config is None:
            if obj.args.server_config == "slurm":
                kwargs = {
                    "account": obj.args.account,
                    "ntasks": obj.args.nproc,
                    "partition": obj.args.partition,
                    "walltime": obj.args.walltime,
                    "mail_type": obj.args.mail_type,
                    "mail_user": obj.args.mail_user,
                    "log_filename": obj.args.log_filename,
                    "modules": obj.args.modules,
                    # "schism_binary": obj.args.schism_binary,
                    "extra_commands": obj.args.extra_commands,
                    "launcher": obj.args.slurm_launcher,
                    "nodes": obj.args.slurm_nodes
                }
                # if obj.args.slurm_filename is not None:
                #     kwargs.update({"filename": obj.args.ntasks})
                # if obj.args.slurm_rundir is not None:
                #     kwargs.update({"run_directory": obj.args.slurm_rundir})
                # if obj.args.run_name is not None:
                #     kwargs.update({"run_name": obj.args.run_name})
                server_config = SlurmConfig(**kwargs)
                obj.__dict__['server_config'] = server_config
        return server_config


class ForecastInit:

    project_directory = ProjectDirectory()
    config_file = ConfigFile()
    coldstart_directory = ColdstartDirectory()
    target_datetime = TargetDatetime()
    coldstart_domain = ColdstartDomain()
    coldstart_driver = ColdstartDriver()
    forecast_interval = ForecastInterval()
    static_files_directory = StaticFilesDirectory()
    hgrid_path = HgridPath()
    vgrid_path = VgridPath()
    fgrid_path = FgridPath()
    tides = TidesDescriptor()
    server_config = ServerConfigDescriptor()

    def __init__(self, args: Namespace):

        """
        TODO: Write a file that inidicates this project is currently being
        accessed by a process. Write the pid of this process to the file.
        May be implemented by a "check_project()" method.
        """
        self._args = args
        self._write_config_file()
        self._symlink_files(self.coldstart_directory,
                            self.coldstart_domain.ics)
        # self.logger.info('Writting coldstart files to disk...')
        self.coldstart_driver.write(
            self.coldstart_directory,
            hgrid=False,
            vgrid=False,
            fgrid=False,
            wind_rot=False,
            overwrite=self.args.overwrite
        )

        if self.args.skip_run is False:
            # release memory before launching SCHISM.
            # self.logger.info('Releasing memory before calling SCHISM...')
            for item in list(self.__dict__.keys()):
                if not item.startswith('_'):
                    del self.__dict__[item]
            # self.logger.info('Calling SCHISM using make.')
            subprocess.check_call(
                ["make", "run"],
                cwd=self.coldstart_directory
            )
        # else:
            # self.logger.info("Skipping coldstart run.")

        # self.logger.info("Finished coldstart sequence.")

    @property
    def args(self):
        return self._args

    def _write_config_file(self):
        if self.config_file.is_file() and self.args.overwrite is False:
            raise IOError(
                f'The given directory {str(self.project_directory)} has '
                'been initialized previously. Please use\npyschism '
                'forecast --overwrite init [...]\nto allow overwrite of '
                'previous initialization options.')
        # self.logger.info(
            # f"Writting configuration file to path {self.config_file}")
        with open(self.config_file, 'w') as fp:
            json.dump(self.args.__dict__, fp, indent=4)

    def _symlink_files(self, target_directory, ics):
        # self.logger.info(
            # f"Establishing symlinks to target_directory: {target_directory}")
        hgrid_lnk = target_directory / 'hgrid.gr3'
        vgrid_lnk = target_directory / 'vgrid.in'
        fgrid_lnk = target_directory / f'{self.fgrid_path.name}'
        if self.args.overwrite is True:
            if hgrid_lnk.exists():
                hgrid_lnk.unlink()
            if vgrid_lnk.exists():
                vgrid_lnk.unlink()
            if fgrid_lnk.exists():
                fgrid_lnk.unlink()
        os.symlink(
            os.path.relpath(
                self.hgrid_path, target_directory), hgrid_lnk)
        os.symlink(
            os.path.relpath(
                self.vgrid_path, target_directory), vgrid_lnk)
        os.symlink(
            os.path.relpath(
                self.fgrid_path, target_directory), fgrid_lnk)
        if ics == 2:
            self._symlink_hgridll(target_directory)

    def _symlink_hgridll(self, target_directory):
        hgridll_lnk = target_directory / 'hgrid.ll'
        if hgridll_lnk.exists():
            hgridll_lnk.unlink()
        if self.args.overwrite is True:
            if self.args.overwrite is True:
                if hgridll_lnk.exists():
                    hgridll_lnk.unlink()
        os.symlink(os.path.relpath(
            self.hgrid_path, target_directory), hgridll_lnk)

    # @property
    # def logger(self):
    #     try:
    #         return self._logger
    #     except AttributeError:
    #         self._logger = get_logger(
    #             console_level=logging._nameToLevel[self.args.log_level.upper()]
    #             )
    #         self._logger.propagate = 0
    #         return self._logger



#! /usr/bin/env python
import unittest
import argparse
from unittest.mock import patch
import sys
from pyschism.cmd import argument_parser


class PlotMeshCmdTestCase(unittest.TestCase):

    def test_add_tidal_run_options(self):
        cmd = ["", "NULL", "start_date", "end_date", "--spinup-days=0"]
        with patch.object(sys, 'argv', cmd):
            self.assertIsInstance(
                argument_parser.get_parser('tidal'),
                argparse.ArgumentParser,
                )

    def test_add_best_track_run_options(self):
        cmd = ["", "NULL", "storm_id", "--start-date", "--end-date"]
        with patch.object(sys, 'argv', cmd):
            self.assertIsInstance(
                argument_parser.get_parser('best_track'),
                argparse.ArgumentParser
                )

    def test_nproc_required(self):
        parser = argparse.ArgumentParser()
        with patch.object(sys, 'argv', ["", "NULL", "--hostname", "NULL"]):
            self.assertIsNone(argument_parser.server(parser))


if __name__ == '__main__':
    unittest.main()



from argparse import Namespace
from datetime import datetime, timedelta
import logging
import pathlib
from time import time

from pyschism.io import Bctides
from pyschism.mesh import Hgrid

_logger = logging.getLogger(__name__)


class BctidesCli:

    def __init__(self, args: Namespace):

        # logging.basicConfig(level=logging._nameToLevel[args.log_level.upper()])
        root_logger = logging.getLogger()
        root_logger.setLevel(logging._nameToLevel[args.log_level.upper()])
        _logger.setLevel(logging._nameToLevel[args.log_level.upper()])

        _logger.info(f'Open hgrid file: {args.hgrid}')
        start = time()
        hgrid = Hgrid.open(args.hgrid, crs=args.hgrid_crs)
        _logger.info(f'Reading hgrid file took {time()-start} seconds.')
        bctides = Bctides(
            hgrid,
            args.start_date,
            args.run_days,
            tidal_database=args.tidal_database,
            velocity=args.include_velocity,
        )

        if args.output_file is not None:
            bctides.write(args.output_file, overwrite=args.overwrite)
        else:
            print(str(bctides))


def add_bctides(subparsers):
    bctides = subparsers.add_parser('bctides')
    bctides.add_argument('hgrid')
    bctides.add_argument(
        'start_date', type=lambda x: datetime.strptime(x, '%Y-%m-%dT%H:%M:%S'))
    bctides.add_argument('run_days', type=float)
    bctides.add_argument('--output-file', type=pathlib.Path)
    bctides.add_argument('--include-velocity', action='store_true')
    bctides.add_argument('--tidal-database', '--tidal-db',
                         choices=['hamtide', 'tpxo'],
                         default='hamtide')
    bctides.add_argument('--hgrid-crs')
    bctides.add_argument(
        "--overwrite", action="store_true",
        help="Allow overwrite of output file.")
    bctides.add_argument(
        "--log-level",
        choices=[name.lower() for name in logging._nameToLevel])
